#
# Copyright (C) 2019 EfficiOS Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; only version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import unittest
import datetime
import bt2
import os
import os.path


_BT_TESTS_DATADIR = os.environ['BT_TESTS_DATADIR']
_BT_CTF_TRACES_PATH = os.environ['BT_CTF_TRACES_PATH']
_3EVENTS_INTERSECT_TRACE_PATH = os.path.join(
    _BT_CTF_TRACES_PATH, 'intersection', '3eventsintersect'
)
_NOINTERSECT_TRACE_PATH = os.path.join(
    _BT_CTF_TRACES_PATH, 'intersection', 'nointersect'
)
_SEQUENCE_TRACE_PATH = os.path.join(_BT_CTF_TRACES_PATH, 'succeed', 'sequence')
_AUTO_SOURCE_DISCOVERY_GROUPING_PATH = os.path.join(
    _BT_TESTS_DATADIR, 'auto-source-discovery', 'grouping'
)
_AUTO_SOURCE_DISCOVERY_PARAMS_LOG_LEVEL_PATH = os.path.join(
    _BT_TESTS_DATADIR, 'auto-source-discovery', 'params-log-level'
)


class _SomeSource(
    bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
):
    pass


class _SomeFilter(
    bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
):
    pass


class _SomeSink(bt2._UserSinkComponent):
    def _user_consume(self):
        pass


class ComponentSpecTestCase(unittest.TestCase):
    def setUp(self):
        # A source CC from a plugin.
        self._dmesg_cc = bt2.find_plugin('text').source_component_classes['dmesg']
        assert self._dmesg_cc is not None

        # A filter CC from a plugin.
        self._muxer_cc = bt2.find_plugin('utils').filter_component_classes['muxer']
        assert self._muxer_cc is not None

        # A sink CC from a plugin.
        self._pretty_cc = bt2.find_plugin('text').sink_component_classes['pretty']
        assert self._pretty_cc is not None

    def test_create_source_from_name(self):
        spec = bt2.ComponentSpec.from_named_plugin_and_component_class('text', 'dmesg')
        self.assertEqual(spec.component_class.name, 'dmesg')

    def test_create_source_from_plugin(self):
        spec = bt2.ComponentSpec(self._dmesg_cc)
        self.assertEqual(spec.component_class.name, 'dmesg')

    def test_create_source_from_user(self):
        spec = bt2.ComponentSpec(_SomeSource)
        self.assertEqual(spec.component_class.name, '_SomeSource')

    def test_create_filter_from_name(self):
        spec = bt2.ComponentSpec.from_named_plugin_and_component_class('utils', 'muxer')
        self.assertEqual(spec.component_class.name, 'muxer')

    def test_create_filter_from_object(self):
        spec = bt2.ComponentSpec(self._muxer_cc)
        self.assertEqual(spec.component_class.name, 'muxer')

    def test_create_sink_from_name(self):
        with self.assertRaisesRegex(
            KeyError,
            'source or filter component class `pretty` not found in plugin `text`',
        ):
            bt2.ComponentSpec.from_named_plugin_and_component_class('text', 'pretty')

    def test_create_sink_from_object(self):
        with self.assertRaisesRegex(
            TypeError,
            "'_SinkComponentClassConst' is not a source or filter component class",
        ):
            bt2.ComponentSpec(self._pretty_cc)

    def test_create_from_object_with_params(self):
        spec = bt2.ComponentSpec(self._dmesg_cc, {'salut': 23})
        self.assertEqual(spec.params['salut'], 23)

    def test_create_from_name_with_params(self):
        spec = bt2.ComponentSpec.from_named_plugin_and_component_class(
            'text', 'dmesg', {'salut': 23}
        )
        self.assertEqual(spec.params['salut'], 23)

    def test_create_from_object_with_path_params(self):
        spec = spec = bt2.ComponentSpec(self._dmesg_cc, 'a path')
        self.assertEqual(spec.params['inputs'], ['a path'])

    def test_create_from_name_with_path_params(self):
        spec = spec = bt2.ComponentSpec.from_named_plugin_and_component_class(
            'text', 'dmesg', 'a path'
        )
        self.assertEqual(spec.params['inputs'], ['a path'])

    def test_create_wrong_comp_class_type(self):
        with self.assertRaisesRegex(
            TypeError, "'int' is not a source or filter component class"
        ):
            bt2.ComponentSpec(18)

    def test_create_from_name_wrong_plugin_name_type(self):
        with self.assertRaisesRegex(TypeError, "'int' is not a 'str' object"):
            bt2.ComponentSpec.from_named_plugin_and_component_class(23, 'compcls')

    def test_create_from_name_non_existent_plugin(self):
        with self.assertRaisesRegex(
            ValueError, "no such plugin: this_plugin_does_not_exist"
        ):
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'this_plugin_does_not_exist', 'compcls'
            )

    def test_create_from_name_wrong_component_class_name_type(self):
        with self.assertRaisesRegex(TypeError, "'int' is not a 'str' object"):
            bt2.ComponentSpec.from_named_plugin_and_component_class('utils', 190)

    def test_create_wrong_params_type(self):
        with self.assertRaisesRegex(
            TypeError, "cannot create value object from 'datetime' object"
        ):
            bt2.ComponentSpec(self._dmesg_cc, params=datetime.datetime.now())

    def test_create_from_name_wrong_params_type(self):
        with self.assertRaisesRegex(
            TypeError, "cannot create value object from 'datetime' object"
        ):
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'text', 'dmesg', datetime.datetime.now()
            )

    def test_create_wrong_log_level_type(self):
        with self.assertRaisesRegex(TypeError, "'str' is not an 'int' object"):
            bt2.ComponentSpec(self._dmesg_cc, logging_level='banane')

    def test_create_from_name_wrong_log_level_type(self):
        with self.assertRaisesRegex(TypeError, "'str' is not an 'int' object"):
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'text', 'dmesg', logging_level='banane'
            )


# Return a map, msg type -> number of messages of this type.


def _count_msgs_by_type(msgs):
    res = {}

    for msg in msgs:
        t = type(msg)
        n = res.get(t, 0)
        res[t] = n + 1

    return res


class TraceCollectionMessageIteratorTestCase(unittest.TestCase):
    def test_create_wrong_stream_intersection_mode_type(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]

        with self.assertRaises(TypeError):
            bt2.TraceCollectionMessageIterator(specs, stream_intersection_mode=23)

    def test_create_wrong_begin_type(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]

        with self.assertRaises(TypeError):
            bt2.TraceCollectionMessageIterator(specs, begin='hi')

    def test_create_wrong_end_type(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]

        with self.assertRaises(TypeError):
            bt2.TraceCollectionMessageIterator(specs, begin='lel')

    def test_create_begin_s(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        bt2.TraceCollectionMessageIterator(specs, begin=19457.918232)

    def test_create_end_s(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        bt2.TraceCollectionMessageIterator(specs, end=123.12312)

    def test_create_begin_datetime(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        bt2.TraceCollectionMessageIterator(specs, begin=datetime.datetime.now())

    def test_create_end_datetime(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        bt2.TraceCollectionMessageIterator(specs, end=datetime.datetime.now())

    def test_iter_no_intersection(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        msg_iter = bt2.TraceCollectionMessageIterator(specs)
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 28)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 8)

    # Same as the above, but we pass a single spec instead of a spec list.
    def test_iter_specs_not_list(self):
        spec = bt2.ComponentSpec.from_named_plugin_and_component_class(
            'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
        )
        msg_iter = bt2.TraceCollectionMessageIterator(spec)
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 28)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 8)

    def test_iter_custom_filter(self):
        src_spec = bt2.ComponentSpec.from_named_plugin_and_component_class(
            'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
        )
        flt_spec = bt2.ComponentSpec.from_named_plugin_and_component_class(
            'utils', 'trimmer', {'end': '13515309.000000075'}
        )
        msg_iter = bt2.TraceCollectionMessageIterator(src_spec, flt_spec)
        hist = _count_msgs_by_type(msg_iter)
        self.assertEqual(hist[bt2._EventMessageConst], 5)

    def test_iter_intersection(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        msg_iter = bt2.TraceCollectionMessageIterator(
            specs, stream_intersection_mode=True
        )
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 15)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 3)

    def test_iter_intersection_params(self):
        # Check that all params used to create the source component are passed
        # to the `babeltrace.trace-infos` query.
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf',
                'fs',
                {
                    'inputs': [_3EVENTS_INTERSECT_TRACE_PATH],
                    'clock-class-offset-s': 1000,
                },
            )
        ]

        msg_iter = bt2.TraceCollectionMessageIterator(
            specs, stream_intersection_mode=True
        )

        event_msgs = [x for x in msg_iter if type(x) is bt2._EventMessageConst]
        self.assertEqual(len(event_msgs), 3)
        self.assertEqual(
            event_msgs[0].default_clock_snapshot.ns_from_origin, 13516309000000071
        )
        self.assertEqual(
            event_msgs[1].default_clock_snapshot.ns_from_origin, 13516309000000072
        )
        self.assertEqual(
            event_msgs[2].default_clock_snapshot.ns_from_origin, 13516309000000082
        )

    def test_iter_no_intersection_two_traces(self):
        spec = bt2.ComponentSpec.from_named_plugin_and_component_class(
            'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
        )
        specs = [spec, spec]
        msg_iter = bt2.TraceCollectionMessageIterator(specs)
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 56)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 16)

    def test_iter_no_intersection_begin(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        msg_iter = bt2.TraceCollectionMessageIterator(specs, begin=13515309.000000023)
        hist = _count_msgs_by_type(msg_iter)
        self.assertEqual(hist[bt2._EventMessageConst], 6)

    def test_iter_no_intersection_end(self):
        specs = [
            bt2.ComponentSpec.from_named_plugin_and_component_class(
                'ctf', 'fs', _3EVENTS_INTERSECT_TRACE_PATH
            )
        ]
        msg_iter = bt2.TraceCollectionMessageIterator(specs, end=13515309.000000075)
        hist = _count_msgs_by_type(msg_iter)
        self.assertEqual(hist[bt2._EventMessageConst], 5)

    def test_iter_auto_source_component_spec(self):
        specs = [bt2.AutoSourceComponentSpec(_3EVENTS_INTERSECT_TRACE_PATH)]
        msg_iter = bt2.TraceCollectionMessageIterator(specs)
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 28)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 8)

    def test_iter_auto_source_component_spec_list_of_strings(self):
        msg_iter = bt2.TraceCollectionMessageIterator([_3EVENTS_INTERSECT_TRACE_PATH])
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 28)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 8)

    def test_iter_auto_source_component_spec_string(self):
        msg_iter = bt2.TraceCollectionMessageIterator(_3EVENTS_INTERSECT_TRACE_PATH)
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 28)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 8)

    def test_iter_mixed_inputs(self):
        msg_iter = bt2.TraceCollectionMessageIterator(
            [
                _3EVENTS_INTERSECT_TRACE_PATH,
                bt2.AutoSourceComponentSpec(_SEQUENCE_TRACE_PATH),
                bt2.ComponentSpec.from_named_plugin_and_component_class(
                    'ctf', 'fs', _NOINTERSECT_TRACE_PATH
                ),
            ]
        )
        msgs = list(msg_iter)
        self.assertEqual(len(msgs), 76)
        hist = _count_msgs_by_type(msgs)
        self.assertEqual(hist[bt2._EventMessageConst], 24)

    def test_auto_source_component_non_existent(self):
        with self.assertRaisesRegex(
            RuntimeError,
            'Some auto source component specs did not produce any component',
        ):
            # Test with one path known to contain a trace and one path known
            # to not contain any trace.
            bt2.TraceCollectionMessageIterator(
                [_SEQUENCE_TRACE_PATH, '/this/path/better/not/exist']
            )


class _TestAutoDiscoverSourceComponentSpecs(unittest.TestCase):
    def setUp(self):
        self._saved_babeltrace_plugin_path = os.environ['BABELTRACE_PLUGIN_PATH']
        os.environ['BABELTRACE_PLUGIN_PATH'] += os.pathsep + self._plugin_path

    def tearDown(self):
        os.environ['BABELTRACE_PLUGIN_PATH'] = self._saved_babeltrace_plugin_path


class TestAutoDiscoverSourceComponentSpecsGrouping(
    _TestAutoDiscoverSourceComponentSpecs
):
    _plugin_path = _AUTO_SOURCE_DISCOVERY_GROUPING_PATH

    def test_grouping(self):
        specs = [
            bt2.AutoSourceComponentSpec('ABCDE'),
            bt2.AutoSourceComponentSpec(_AUTO_SOURCE_DISCOVERY_GROUPING_PATH),
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 8)

        self.assertEqual(msgs[0].stream.name, 'TestSourceABCDE: ABCDE')
        self.assertEqual(msgs[1].stream.name, 'TestSourceExt: aaa1, aaa2, aaa3')
        self.assertEqual(msgs[2].stream.name, 'TestSourceExt: bbb1, bbb2')
        self.assertEqual(msgs[3].stream.name, 'TestSourceExt: ccc1')
        self.assertEqual(msgs[4].stream.name, 'TestSourceExt: ccc2')
        self.assertEqual(msgs[5].stream.name, 'TestSourceExt: ccc3')
        self.assertEqual(msgs[6].stream.name, 'TestSourceExt: ccc4')
        self.assertEqual(msgs[7].stream.name, 'TestSourceSomeDir: some-dir')


class TestAutoDiscoverSourceComponentSpecsParamsObjLogLevel(
    _TestAutoDiscoverSourceComponentSpecs
):
    _plugin_path = _AUTO_SOURCE_DISCOVERY_PARAMS_LOG_LEVEL_PATH

    _dir_a = os.path.join(_AUTO_SOURCE_DISCOVERY_PARAMS_LOG_LEVEL_PATH, 'dir-a')
    _dir_b = os.path.join(_AUTO_SOURCE_DISCOVERY_PARAMS_LOG_LEVEL_PATH, 'dir-b')
    _dir_ab = os.path.join(_AUTO_SOURCE_DISCOVERY_PARAMS_LOG_LEVEL_PATH, 'dir-ab')

    def _test_two_comps_from_one_spec(self, params, obj=None, logging_level=None):
        specs = [
            bt2.AutoSourceComponentSpec(
                self._dir_ab, params=params, obj=obj, logging_level=logging_level
            )
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 2)

        return msgs

    def test_params_two_comps_from_one_spec(self):
        msgs = self._test_two_comps_from_one_spec(
            params={'test-allo': 'madame', 'what': 'test-params'}
        )

        self.assertEqual(msgs[0].stream.name, "TestSourceA: ('test-allo', 'madame')")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: ('test-allo', 'madame')")

    def test_obj_two_comps_from_one_spec(self):
        msgs = self._test_two_comps_from_one_spec(
            params={'what': 'python-obj'}, obj='deore'
        )

        self.assertEqual(msgs[0].stream.name, "TestSourceA: deore")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: deore")

    def test_log_level_two_comps_from_one_spec(self):
        msgs = self._test_two_comps_from_one_spec(
            params={'what': 'log-level'}, logging_level=bt2.LoggingLevel.DEBUG
        )

        self.assertEqual(
            msgs[0].stream.name, "TestSourceA: {}".format(bt2.LoggingLevel.DEBUG)
        )
        self.assertEqual(
            msgs[1].stream.name, "TestSourceB: {}".format(bt2.LoggingLevel.DEBUG)
        )

    def _test_two_comps_from_two_specs(
        self,
        params_a=None,
        params_b=None,
        obj_a=None,
        obj_b=None,
        logging_level_a=None,
        logging_level_b=None,
    ):
        specs = [
            bt2.AutoSourceComponentSpec(
                self._dir_a, params=params_a, obj=obj_a, logging_level=logging_level_a
            ),
            bt2.AutoSourceComponentSpec(
                self._dir_b, params=params_b, obj=obj_b, logging_level=logging_level_b
            ),
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 2)

        return msgs

    def test_params_two_comps_from_two_specs(self):
        msgs = self._test_two_comps_from_two_specs(
            params_a={'test-allo': 'madame', 'what': 'test-params'},
            params_b={'test-bonjour': 'monsieur', 'what': 'test-params'},
        )

        self.assertEqual(msgs[0].stream.name, "TestSourceA: ('test-allo', 'madame')")
        self.assertEqual(
            msgs[1].stream.name, "TestSourceB: ('test-bonjour', 'monsieur')"
        )

    def test_obj_two_comps_from_two_specs(self):
        msgs = self._test_two_comps_from_two_specs(
            params_a={'what': 'python-obj'},
            params_b={'what': 'python-obj'},
            obj_a='deore',
            obj_b='alivio',
        )

        self.assertEqual(msgs[0].stream.name, "TestSourceA: deore")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: alivio")

    def test_log_level_two_comps_from_two_specs(self):
        msgs = self._test_two_comps_from_two_specs(
            params_a={'what': 'log-level'},
            params_b={'what': 'log-level'},
            logging_level_a=bt2.LoggingLevel.DEBUG,
            logging_level_b=bt2.LoggingLevel.TRACE,
        )

        self.assertEqual(
            msgs[0].stream.name, "TestSourceA: {}".format(bt2.LoggingLevel.DEBUG)
        )
        self.assertEqual(
            msgs[1].stream.name, "TestSourceB: {}".format(bt2.LoggingLevel.TRACE)
        )

    def _test_one_comp_from_one_spec_one_comp_from_both_1(
        self,
        params_a=None,
        params_ab=None,
        obj_a=None,
        obj_ab=None,
        logging_level_a=None,
        logging_level_ab=None,
    ):
        specs = [
            bt2.AutoSourceComponentSpec(
                self._dir_a, params=params_a, obj=obj_a, logging_level=logging_level_a
            ),
            bt2.AutoSourceComponentSpec(
                self._dir_ab,
                params=params_ab,
                obj=obj_ab,
                logging_level=logging_level_ab,
            ),
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 2)

        return msgs

    def test_params_one_comp_from_one_spec_one_comp_from_both_1(self):
        msgs = self._test_one_comp_from_one_spec_one_comp_from_both_1(
            params_a={'test-allo': 'madame', 'what': 'test-params'},
            params_ab={'test-bonjour': 'monsieur', 'what': 'test-params'},
        )

        self.assertEqual(
            msgs[0].stream.name,
            "TestSourceA: ('test-allo', 'madame'), ('test-bonjour', 'monsieur')",
        )
        self.assertEqual(
            msgs[1].stream.name, "TestSourceB: ('test-bonjour', 'monsieur')"
        )

    def test_obj_one_comp_from_one_spec_one_comp_from_both_1(self):
        msgs = self._test_one_comp_from_one_spec_one_comp_from_both_1(
            params_a={'what': 'python-obj'},
            params_ab={'what': 'python-obj'},
            obj_a='deore',
            obj_ab='alivio',
        )

        self.assertEqual(msgs[0].stream.name, "TestSourceA: alivio")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: alivio")

    def test_log_level_one_comp_from_one_spec_one_comp_from_both_1(self):
        msgs = self._test_one_comp_from_one_spec_one_comp_from_both_1(
            params_a={'what': 'log-level'},
            params_ab={'what': 'log-level'},
            logging_level_a=bt2.LoggingLevel.DEBUG,
            logging_level_ab=bt2.LoggingLevel.TRACE,
        )

        self.assertEqual(
            msgs[0].stream.name, "TestSourceA: {}".format(bt2.LoggingLevel.TRACE)
        )
        self.assertEqual(
            msgs[1].stream.name, "TestSourceB: {}".format(bt2.LoggingLevel.TRACE)
        )

    def _test_one_comp_from_one_spec_one_comp_from_both_2(
        self,
        params_ab=None,
        params_a=None,
        obj_ab=None,
        obj_a=None,
        logging_level_ab=None,
        logging_level_a=None,
    ):
        specs = [
            bt2.AutoSourceComponentSpec(
                self._dir_ab,
                params=params_ab,
                obj=obj_ab,
                logging_level=logging_level_ab,
            ),
            bt2.AutoSourceComponentSpec(
                self._dir_a, params=params_a, obj=obj_a, logging_level=logging_level_a
            ),
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 2)

        return msgs

    def test_params_one_comp_from_one_spec_one_comp_from_both_2(self):
        msgs = self._test_one_comp_from_one_spec_one_comp_from_both_2(
            params_ab={
                'test-bonjour': 'madame',
                'test-salut': 'les amis',
                'what': 'test-params',
            },
            params_a={'test-bonjour': 'monsieur', 'what': 'test-params'},
        )

        self.assertEqual(
            msgs[0].stream.name,
            "TestSourceA: ('test-bonjour', 'monsieur'), ('test-salut', 'les amis')",
        )
        self.assertEqual(
            msgs[1].stream.name,
            "TestSourceB: ('test-bonjour', 'madame'), ('test-salut', 'les amis')",
        )

    def test_obj_one_comp_from_one_spec_one_comp_from_both_2(self):
        msgs = self._test_one_comp_from_one_spec_one_comp_from_both_2(
            params_ab={'what': 'python-obj'},
            params_a={'what': 'python-obj'},
            obj_ab='deore',
            obj_a='alivio',
        )

        self.assertEqual(msgs[0].stream.name, "TestSourceA: alivio")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: deore")

    def test_log_level_one_comp_from_one_spec_one_comp_from_both_2(self):
        msgs = self._test_one_comp_from_one_spec_one_comp_from_both_2(
            params_ab={'what': 'log-level'},
            params_a={'what': 'log-level'},
            logging_level_ab=bt2.LoggingLevel.DEBUG,
            logging_level_a=bt2.LoggingLevel.TRACE,
        )

        self.assertEqual(
            msgs[0].stream.name, "TestSourceA: {}".format(bt2.LoggingLevel.TRACE)
        )
        self.assertEqual(
            msgs[1].stream.name, "TestSourceB: {}".format(bt2.LoggingLevel.DEBUG)
        )

    def test_obj_override_with_none(self):
        specs = [
            bt2.AutoSourceComponentSpec(
                self._dir_ab, params={'what': 'python-obj'}, obj='deore'
            ),
            bt2.AutoSourceComponentSpec(
                self._dir_a, params={'what': 'python-obj'}, obj=None
            ),
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 2)
        self.assertEqual(msgs[0].stream.name, "TestSourceA: None")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: deore")

    def test_obj_no_override_with_no_obj(self):
        specs = [
            bt2.AutoSourceComponentSpec(
                self._dir_ab, params={'what': 'python-obj'}, obj='deore'
            ),
            bt2.AutoSourceComponentSpec(self._dir_a, params={'what': 'python-obj'}),
        ]
        it = bt2.TraceCollectionMessageIterator(specs)
        msgs = [x for x in it if type(x) is bt2._StreamBeginningMessageConst]

        self.assertEqual(len(msgs), 2)
        self.assertEqual(msgs[0].stream.name, "TestSourceA: deore")
        self.assertEqual(msgs[1].stream.name, "TestSourceB: deore")


if __name__ == '__main__':
    unittest.main()



import bt2
import os

# This file defines source component classes to help verify the parameters an
# log levels passed to components.  Each component creates one stream, with a
# name derived from either:
#
#   - the received params that start with `test-`
#   - the received log level
#
# The `what` parameter determines what is used.


class TestIter(bt2._UserMessageIterator):
    def __init__(self, config, output_port):
        params = output_port.user_data['params']
        obj = output_port.user_data['obj']

        comp_cls_name = self._component.__class__.__name__

        if params['what'] == 'test-params':
            items = sorted([str(x) for x in params.items() if x[0].startswith('test-')])
            stream_name = '{}: {}'.format(comp_cls_name, ', '.join(items))
        elif params['what'] == 'log-level':
            log_level = self._component.logging_level
            stream_name = '{}: {}'.format(comp_cls_name, log_level)
        elif params['what'] == 'python-obj':
            assert type(obj) == str or obj is None
            stream_name = '{}: {}'.format(comp_cls_name, obj)
        else:
            assert False

        sc = output_port.user_data['sc']
        tc = sc.trace_class
        t = tc()
        s = t.create_stream(sc, name=stream_name)

        self._msgs = [
            self._create_stream_beginning_message(s),
            self._create_stream_end_message(s),
        ]

    def __next__(self):
        if len(self._msgs) == 0:
            raise StopIteration

        return self._msgs.pop(0)


class Base:
    def __init__(self, params, obj):
        tc = self._create_trace_class()
        sc = tc.create_stream_class()

        self._add_output_port('out', {'params': params, 'obj': obj, 'sc': sc})


@bt2.plugin_component_class
class TestSourceA(Base, bt2._UserSourceComponent, message_iterator_class=TestIter):
    def __init__(self, config, params, obj):
        super().__init__(params, obj)

    @staticmethod
    def _user_query(priv_query_exec, obj, params, method_obj):
        # Match files starting with 'aaa'.

        if obj == 'babeltrace.support-info':
            if params['type'] != 'file':
                return 0

            name = os.path.basename(str(params['input']))

            if name.startswith('aaa'):
                return {'weight': 1, 'group': 'aaa'}
            else:
                return 0
        else:
            raise bt2.UnknownObject


@bt2.plugin_component_class
class TestSourceB(Base, bt2._UserSourceComponent, message_iterator_class=TestIter):
    def __init__(self, config, params, obj):
        super().__init__(params, obj)

    @staticmethod
    def _user_query(priv_query_exec, obj, params, method_obj):
        # Match files starting with 'bbb'.

        if obj == 'babeltrace.support-info':
            if params['type'] != 'file':
                return 0

            name = os.path.basename(str(params['input']))

            if name.startswith('bbb'):
                return {'weight': 1, 'group': 'bbb'}
            else:
                return 0
        else:
            raise bt2.UnknownObject


bt2.register_plugin(module_name=__name__, name="test")



# Copyright (c) 2016, Matt Layman

import os
import unittest

from tap.adapter import Adapter
from tap.parser import Parser
from tap.rules import Rules


class Loader(object):
    """Load TAP lines into unittest-able objects."""

    ignored_lines = set(['diagnostic', 'unknown'])

    def __init__(self):
        self._parser = Parser()

    def load(self, files):
        """Load any files found into a suite.

        Any directories are walked and their files are added as TAP files.

        :returns: A ``unittest.TestSuite`` instance
        """
        suite = unittest.TestSuite()
        for filepath in files:
            if os.path.isdir(filepath):
                self._find_tests_in_directory(filepath, suite)
            else:
                suite.addTest(self.load_suite_from_file(filepath))
        return suite

    def load_suite_from_file(self, filename):
        """Load a test suite with test lines from the provided TAP file.

        :returns: A ``unittest.TestSuite`` instance
        """
        suite = unittest.TestSuite()
        rules = Rules(filename, suite)

        if not os.path.exists(filename):
            rules.handle_file_does_not_exist()
            return suite

        line_generator = self._parser.parse_file(filename)
        return self._load_lines(filename, line_generator, suite, rules)

    def load_suite_from_stdin(self):
        """Load a test suite with test lines from the TAP stream on STDIN.

        :returns: A ``unittest.TestSuite`` instance
        """
        suite = unittest.TestSuite()
        rules = Rules('stream', suite)
        line_generator = self._parser.parse_stdin()
        return self._load_lines('stream', line_generator, suite, rules)

    def _find_tests_in_directory(self, directory, suite):
        """Find test files in the directory and add them to the suite."""
        for dirpath, dirnames, filenames in os.walk(directory):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                suite.addTest(self.load_suite_from_file(filepath))

    def _load_lines(self, filename, line_generator, suite, rules):
        """Load a suite with lines produced by the line generator."""
        line_counter = 0
        for line in line_generator:
            line_counter += 1

            if line.category in self.ignored_lines:
                continue

            if line.category == 'test':
                suite.addTest(Adapter(filename, line))
                rules.saw_test()
            elif line.category == 'plan':
                if line.skip:
                    rules.handle_skipping_plan(line)
                    return suite
                rules.saw_plan(line, line_counter)
            elif line.category == 'bail':
                rules.handle_bail(line)
                return suite
            elif line.category == 'version':
                rules.saw_version_at(line_counter)

        rules.check(line_counter)
        return suite



# The MIT License (MIT)
#
# Copyright (c) 2019 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import argparse
import collections.abc
import logging
import os
import os.path
import re
import socket
import struct
import sys
import tempfile


class UnexpectedInput(RuntimeError):
    pass


class _LttngLiveViewerCommand:
    def __init__(self, version):
        self._version = version

    @property
    def version(self):
        return self._version


class _LttngLiveViewerConnectCommand(_LttngLiveViewerCommand):
    def __init__(self, version, viewer_session_id, major, minor):
        super().__init__(version)
        self._viewer_session_id = viewer_session_id
        self._major = major
        self._minor = minor

    @property
    def viewer_session_id(self):
        return self._viewer_session_id

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor


class _LttngLiveViewerConnectReply:
    def __init__(self, viewer_session_id, major, minor):
        self._viewer_session_id = viewer_session_id
        self._major = major
        self._minor = minor

    @property
    def viewer_session_id(self):
        return self._viewer_session_id

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor


class _LttngLiveViewerGetTracingSessionInfosCommand(_LttngLiveViewerCommand):
    pass


class _LttngLiveViewerTracingSessionInfo:
    def __init__(
        self,
        tracing_session_id,
        live_timer_freq,
        client_count,
        stream_count,
        hostname,
        name,
    ):
        self._tracing_session_id = tracing_session_id
        self._live_timer_freq = live_timer_freq
        self._client_count = client_count
        self._stream_count = stream_count
        self._hostname = hostname
        self._name = name

    @property
    def tracing_session_id(self):
        return self._tracing_session_id

    @property
    def live_timer_freq(self):
        return self._live_timer_freq

    @property
    def client_count(self):
        return self._client_count

    @property
    def stream_count(self):
        return self._stream_count

    @property
    def hostname(self):
        return self._hostname

    @property
    def name(self):
        return self._name


class _LttngLiveViewerGetTracingSessionInfosReply:
    def __init__(self, tracing_session_infos):
        self._tracing_session_infos = tracing_session_infos

    @property
    def tracing_session_infos(self):
        return self._tracing_session_infos


class _LttngLiveViewerAttachToTracingSessionCommand(_LttngLiveViewerCommand):
    class SeekType:
        BEGINNING = 1
        LAST = 2

    def __init__(self, version, tracing_session_id, offset, seek_type):
        super().__init__(version)
        self._tracing_session_id = tracing_session_id
        self._offset = offset
        self._seek_type = seek_type

    @property
    def tracing_session_id(self):
        return self._tracing_session_id

    @property
    def offset(self):
        return self._offset

    @property
    def seek_type(self):
        return self._seek_type


class _LttngLiveViewerStreamInfo:
    def __init__(self, id, trace_id, is_metadata, path, channel_name):
        self._id = id
        self._trace_id = trace_id
        self._is_metadata = is_metadata
        self._path = path
        self._channel_name = channel_name

    @property
    def id(self):
        return self._id

    @property
    def trace_id(self):
        return self._trace_id

    @property
    def is_metadata(self):
        return self._is_metadata

    @property
    def path(self):
        return self._path

    @property
    def channel_name(self):
        return self._channel_name


class _LttngLiveViewerAttachToTracingSessionReply:
    class Status:
        OK = 1
        ALREADY = 2
        UNKNOWN = 3
        NOT_LIVE = 4
        SEEK_ERROR = 5
        NO_SESSION = 6

    def __init__(self, status, stream_infos):
        self._status = status
        self._stream_infos = stream_infos

    @property
    def status(self):
        return self._status

    @property
    def stream_infos(self):
        return self._stream_infos


class _LttngLiveViewerGetNextDataStreamIndexEntryCommand(_LttngLiveViewerCommand):
    def __init__(self, version, stream_id):
        super().__init__(version)
        self._stream_id = stream_id

    @property
    def stream_id(self):
        return self._stream_id


class _LttngLiveViewerGetNextDataStreamIndexEntryReply:
    class Status:
        OK = 1
        RETRY = 2
        HUP = 3
        ERROR = 4
        INACTIVE = 5
        EOF = 6

    def __init__(self, status, index_entry, has_new_metadata, has_new_data_stream):
        self._status = status
        self._index_entry = index_entry
        self._has_new_metadata = has_new_metadata
        self._has_new_data_stream = has_new_data_stream

    @property
    def status(self):
        return self._status

    @property
    def index_entry(self):
        return self._index_entry

    @property
    def has_new_metadata(self):
        return self._has_new_metadata

    @property
    def has_new_data_stream(self):
        return self._has_new_data_stream


class _LttngLiveViewerGetDataStreamPacketDataCommand(_LttngLiveViewerCommand):
    def __init__(self, version, stream_id, offset, req_length):
        super().__init__(version)
        self._stream_id = stream_id
        self._offset = offset
        self._req_length = req_length

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def offset(self):
        return self._offset

    @property
    def req_length(self):
        return self._req_length


class _LttngLiveViewerGetDataStreamPacketDataReply:
    class Status:
        OK = 1
        RETRY = 2
        ERROR = 3
        EOF = 4

    def __init__(self, status, data, has_new_metadata, has_new_data_stream):
        self._status = status
        self._data = data
        self._has_new_metadata = has_new_metadata
        self._has_new_data_stream = has_new_data_stream

    @property
    def status(self):
        return self._status

    @property
    def data(self):
        return self._data

    @property
    def has_new_metadata(self):
        return self._has_new_metadata

    @property
    def has_new_data_stream(self):
        return self._has_new_data_stream


class _LttngLiveViewerGetMetadataStreamDataCommand(_LttngLiveViewerCommand):
    def __init__(self, version, stream_id):
        super().__init__(version)
        self._stream_id = stream_id

    @property
    def stream_id(self):
        return self._stream_id


class _LttngLiveViewerGetMetadataStreamDataContentReply:
    class Status:
        OK = 1
        NO_NEW = 2
        ERROR = 3

    def __init__(self, status, data):
        self._status = status
        self._data = data

    @property
    def status(self):
        return self._status

    @property
    def data(self):
        return self._data


class _LttngLiveViewerGetNewStreamInfosCommand(_LttngLiveViewerCommand):
    def __init__(self, version, tracing_session_id):
        super().__init__(version)
        self._tracing_session_id = tracing_session_id

    @property
    def tracing_session_id(self):
        return self._tracing_session_id


class _LttngLiveViewerGetNewStreamInfosReply:
    class Status:
        OK = 1
        NO_NEW = 2
        ERROR = 3
        HUP = 4

    def __init__(self, status, stream_infos):
        self._status = status
        self._stream_infos = stream_infos

    @property
    def status(self):
        return self._status

    @property
    def stream_infos(self):
        return self._stream_infos


class _LttngLiveViewerCreateViewerSessionCommand(_LttngLiveViewerCommand):
    pass


class _LttngLiveViewerCreateViewerSessionReply:
    class Status:
        OK = 1
        ERROR = 2

    def __init__(self, status):
        self._status = status

    @property
    def status(self):
        return self._status


class _LttngLiveViewerDetachFromTracingSessionCommand(_LttngLiveViewerCommand):
    def __init__(self, version, tracing_session_id):
        super().__init__(version)
        self._tracing_session_id = tracing_session_id

    @property
    def tracing_session_id(self):
        return self._tracing_session_id


class _LttngLiveViewerDetachFromTracingSessionReply:
    class Status:
        OK = 1
        UNKNOWN = 2
        ERROR = 3

    def __init__(self, status):
        self._status = status

    @property
    def status(self):
        return self._status


# An LTTng live protocol codec can convert bytes to command objects and
# reply objects to bytes.
class _LttngLiveViewerProtocolCodec:
    _COMMAND_HEADER_STRUCT_FMT = 'QII'
    _COMMAND_HEADER_SIZE_BYTES = struct.calcsize(_COMMAND_HEADER_STRUCT_FMT)

    def __init__(self):
        pass

    def _unpack(self, fmt, data, offset=0):
        fmt = '!' + fmt
        return struct.unpack_from(fmt, data, offset)

    def _unpack_payload(self, fmt, data):
        return self._unpack(
            fmt, data, _LttngLiveViewerProtocolCodec._COMMAND_HEADER_SIZE_BYTES
        )

    def decode(self, data):
        if len(data) < self._COMMAND_HEADER_SIZE_BYTES:
            # Not enough data to read the command header
            return

        payload_size, cmd_type, version = self._unpack(
            self._COMMAND_HEADER_STRUCT_FMT, data
        )
        logging.info(
            'Decoded command header: payload-size={}, cmd-type={}, version={}'.format(
                payload_size, cmd_type, version
            )
        )

        if len(data) < self._COMMAND_HEADER_SIZE_BYTES + payload_size:
            # Not enough data to read the whole command
            return

        if cmd_type == 1:
            viewer_session_id, major, minor, conn_type = self._unpack_payload(
                'QIII', data
            )
            return _LttngLiveViewerConnectCommand(
                version, viewer_session_id, major, minor
            )
        elif cmd_type == 2:
            return _LttngLiveViewerGetTracingSessionInfosCommand(version)
        elif cmd_type == 3:
            tracing_session_id, offset, seek_type = self._unpack_payload('QQI', data)
            return _LttngLiveViewerAttachToTracingSessionCommand(
                version, tracing_session_id, offset, seek_type
            )
        elif cmd_type == 4:
            (stream_id,) = self._unpack_payload('Q', data)
            return _LttngLiveViewerGetNextDataStreamIndexEntryCommand(
                version, stream_id
            )
        elif cmd_type == 5:
            stream_id, offset, req_length = self._unpack_payload('QQI', data)
            return _LttngLiveViewerGetDataStreamPacketDataCommand(
                version, stream_id, offset, req_length
            )
        elif cmd_type == 6:
            (stream_id,) = self._unpack_payload('Q', data)
            return _LttngLiveViewerGetMetadataStreamDataCommand(version, stream_id)
        elif cmd_type == 7:
            (tracing_session_id,) = self._unpack_payload('Q', data)
            return _LttngLiveViewerGetNewStreamInfosCommand(version, tracing_session_id)
        elif cmd_type == 8:
            return _LttngLiveViewerCreateViewerSessionCommand(version)
        elif cmd_type == 9:
            (tracing_session_id,) = self._unpack_payload('Q', data)
            return _LttngLiveViewerDetachFromTracingSessionCommand(
                version, tracing_session_id
            )
        else:
            raise UnexpectedInput('Unknown command type {}'.format(cmd_type))

    def _pack(self, fmt, *args):
        # Force network byte order
        return struct.pack('!' + fmt, *args)

    def _encode_zero_padded_str(self, string, length):
        data = string.encode()
        return data.ljust(length, b'\x00')

    def _encode_stream_info(self, info):
        data = self._pack('QQI', info.id, info.trace_id, int(info.is_metadata))
        data += self._encode_zero_padded_str(info.path, 4096)
        data += self._encode_zero_padded_str(info.channel_name, 255)
        return data

    def _get_has_new_stuff_flags(self, has_new_metadata, has_new_data_streams):
        flags = 0

        if has_new_metadata:
            flags |= 1

        if has_new_data_streams:
            flags |= 2

        return flags

    def encode(self, reply):
        if type(reply) is _LttngLiveViewerConnectReply:
            data = self._pack(
                'QIII', reply.viewer_session_id, reply.major, reply.minor, 2
            )
        elif type(reply) is _LttngLiveViewerGetTracingSessionInfosReply:
            data = self._pack('I', len(reply.tracing_session_infos))

            for info in reply.tracing_session_infos:
                data += self._pack(
                    'QIII',
                    info.tracing_session_id,
                    info.live_timer_freq,
                    info.client_count,
                    info.stream_count,
                )
                data += self._encode_zero_padded_str(info.hostname, 64)
                data += self._encode_zero_padded_str(info.name, 255)
        elif type(reply) is _LttngLiveViewerAttachToTracingSessionReply:
            data = self._pack('II', reply.status, len(reply.stream_infos))

            for info in reply.stream_infos:
                data += self._encode_stream_info(info)
        elif type(reply) is _LttngLiveViewerGetNextDataStreamIndexEntryReply:
            entry = reply.index_entry
            flags = self._get_has_new_stuff_flags(
                reply.has_new_metadata, reply.has_new_data_stream
            )

            data = self._pack(
                'QQQQQQQII',
                entry.offset_bytes,
                entry.total_size_bits,
                entry.content_size_bits,
                entry.timestamp_begin,
                entry.timestamp_end,
                entry.events_discarded,
                entry.stream_class_id,
                reply.status,
                flags,
            )
        elif type(reply) is _LttngLiveViewerGetDataStreamPacketDataReply:
            flags = self._get_has_new_stuff_flags(
                reply.has_new_metadata, reply.has_new_data_stream
            )
            data = self._pack('III', reply.status, len(reply.data), flags)
            data += reply.data
        elif type(reply) is _LttngLiveViewerGetMetadataStreamDataContentReply:
            data = self._pack('QI', len(reply.data), reply.status)
            data += reply.data
        elif type(reply) is _LttngLiveViewerGetNewStreamInfosReply:
            data = self._pack('II', reply.status, len(reply.stream_infos))

            for info in reply.stream_infos:
                data += self._encode_stream_info(info)
        elif type(reply) is _LttngLiveViewerCreateViewerSessionReply:
            data = self._pack('I', reply.status)
        elif type(reply) is _LttngLiveViewerDetachFromTracingSessionReply:
            data = self._pack('I', reply.status)
        else:
            raise ValueError(
                'Unknown reply object with class `{}`'.format(reply.__class__.__name__)
            )

        return data


# An entry within the index of an LTTng data stream.
class _LttngDataStreamIndexEntry:
    def __init__(
        self,
        offset_bytes,
        total_size_bits,
        content_size_bits,
        timestamp_begin,
        timestamp_end,
        events_discarded,
        stream_class_id,
    ):
        self._offset_bytes = offset_bytes
        self._total_size_bits = total_size_bits
        self._content_size_bits = content_size_bits
        self._timestamp_begin = timestamp_begin
        self._timestamp_end = timestamp_end
        self._events_discarded = events_discarded
        self._stream_class_id = stream_class_id

    @property
    def offset_bytes(self):
        return self._offset_bytes

    @property
    def total_size_bits(self):
        return self._total_size_bits

    @property
    def total_size_bytes(self):
        return self._total_size_bits // 8

    @property
    def content_size_bits(self):
        return self._content_size_bits

    @property
    def content_size_bytes(self):
        return self._content_size_bits // 8

    @property
    def timestamp_begin(self):
        return self._timestamp_begin

    @property
    def timestamp_end(self):
        return self._timestamp_end

    @property
    def events_discarded(self):
        return self._events_discarded

    @property
    def stream_class_id(self):
        return self._stream_class_id


# The index of an LTTng data stream, a sequence of index entries.
class _LttngDataStreamIndex(collections.abc.Sequence):
    def __init__(self, path):
        self._path = path
        self._build()
        logging.info(
            'Built data stream index entries: path="{}", count={}'.format(
                path, len(self._entries)
            )
        )

    def _build(self):
        self._entries = []
        assert os.path.isfile(self._path)

        with open(self._path, 'rb') as f:
            # Read header first
            fmt = '>IIII'
            size = struct.calcsize(fmt)
            data = f.read(size)
            assert len(data) == size
            magic, index_major, index_minor, index_entry_length = struct.unpack(
                fmt, data
            )
            assert magic == 0xC1F1DCC1

            # Read index entries
            fmt = '>QQQQQQQ'
            size = struct.calcsize(fmt)

            while True:
                logging.debug(
                    'Decoding data stream index entry: path="{}", offset={}'.format(
                        self._path, f.tell()
                    )
                )
                data = f.read(size)

                if not data:
                    # Done
                    break

                assert len(data) == size
                (
                    offset_bytes,
                    total_size_bits,
                    content_size_bits,
                    timestamp_begin,
                    timestamp_end,
                    events_discarded,
                    stream_class_id,
                ) = struct.unpack(fmt, data)

                self._entries.append(
                    _LttngDataStreamIndexEntry(
                        offset_bytes,
                        total_size_bits,
                        content_size_bits,
                        timestamp_begin,
                        timestamp_end,
                        events_discarded,
                        stream_class_id,
                    )
                )

                # Skip anything else before the next entry
                f.seek(index_entry_length - size, os.SEEK_CUR)

    def __getitem__(self, index):
        return self._entries[index]

    def __len__(self):
        return len(self._entries)

    @property
    def path(self):
        return self._path


# An LTTng data stream.
class _LttngDataStream:
    def __init__(self, path):
        self._path = path
        filename = os.path.basename(path)
        match = re.match(r'(.*)_\d+', filename)
        self._channel_name = match.group(1)
        trace_dir = os.path.dirname(path)
        index_path = os.path.join(trace_dir, 'index', filename + '.idx')
        self._index = _LttngDataStreamIndex(index_path)
        assert os.path.isfile(path)
        self._file = open(path, 'rb')
        logging.info(
            'Built data stream: path="{}", channel-name="{}"'.format(
                path, self._channel_name
            )
        )

    @property
    def path(self):
        return self._path

    @property
    def channel_name(self):
        return self._channel_name

    @property
    def index(self):
        return self._index

    def get_data(self, offset_bytes, len_bytes):
        self._file.seek(offset_bytes)
        return self._file.read(len_bytes)


# An LTTng metadata stream.
class _LttngMetadataStream:
    def __init__(self, path):
        self._path = path
        logging.info('Built metadata stream: path="{}"'.format(path))

    @property
    def path(self):
        return self._path

    @property
    def data(self):
        assert os.path.isfile(self._path)

        with open(self._path, 'rb') as f:
            return f.read()


# An LTTng trace, a sequence of LTTng data streams.
class LttngTrace(collections.abc.Sequence):
    def __init__(self, trace_dir):
        assert os.path.isdir(trace_dir)
        self._path = trace_dir
        self._metadata_stream = _LttngMetadataStream(
            os.path.join(trace_dir, 'metadata')
        )
        self._create_data_streams(trace_dir)
        logging.info('Built trace: path="{}"'.format(trace_dir))

    def _create_data_streams(self, trace_dir):
        data_stream_paths = []

        for filename in os.listdir(trace_dir):
            path = os.path.join(trace_dir, filename)

            if not os.path.isfile(path):
                continue

            if filename.startswith('.'):
                continue

            if filename == 'metadata':
                continue

            data_stream_paths.append(path)

        data_stream_paths.sort()
        self._data_streams = []

        for data_stream_path in data_stream_paths:
            self._data_streams.append(_LttngDataStream(data_stream_path))

    @property
    def path(self):
        return self._path

    @property
    def metadata_stream(self):
        return self._metadata_stream

    def __getitem__(self, index):
        return self._data_streams[index]

    def __len__(self):
        return len(self._data_streams)


# The state of a single data stream.
class _LttngLiveViewerSessionDataStreamState:
    def __init__(self, ts_state, info, data_stream):
        self._ts_state = ts_state
        self._info = info
        self._data_stream = data_stream
        self._cur_index_entry_index = 0
        fmt = 'Built data stream state: id={}, ts-id={}, ts-name="{}", path="{}"'
        logging.info(
            fmt.format(
                info.id,
                ts_state.tracing_session_descriptor.info.tracing_session_id,
                ts_state.tracing_session_descriptor.info.name,
                data_stream.path,
            )
        )

    @property
    def tracing_session_state(self):
        return self._ts_state

    @property
    def info(self):
        return self._info

    @property
    def data_stream(self):
        return self._data_stream

    @property
    def cur_index_entry(self):
        if self._cur_index_entry_index == len(self._data_stream.index):
            return

        return self._data_stream.index[self._cur_index_entry_index]

    def goto_next_index_entry(self):
        self._cur_index_entry_index += 1


# The state of a single metadata stream.
class _LttngLiveViewerSessionMetadataStreamState:
    def __init__(self, ts_state, info, metadata_stream):
        self._ts_state = ts_state
        self._info = info
        self._metadata_stream = metadata_stream
        self._is_sent = False
        fmt = 'Built metadata stream state: id={}, ts-id={}, ts-name="{}", path="{}"'
        logging.info(
            fmt.format(
                info.id,
                ts_state.tracing_session_descriptor.info.tracing_session_id,
                ts_state.tracing_session_descriptor.info.name,
                metadata_stream.path,
            )
        )

    @property
    def trace_session_state(self):
        return self._trace_session_state

    @property
    def info(self):
        return self._info

    @property
    def metadata_stream(self):
        return self._metadata_stream

    @property
    def is_sent(self):
        return self._is_sent

    @is_sent.setter
    def is_sent(self, value):
        self._is_sent = value


# The state of a tracing session.
class _LttngLiveViewerSessionTracingSessionState:
    def __init__(self, tc_descr, base_stream_id):
        self._tc_descr = tc_descr
        self._stream_infos = []
        self._ds_states = {}
        self._ms_states = {}
        stream_id = base_stream_id

        for trace in tc_descr.traces:
            trace_id = stream_id * 1000

            # Data streams -> stream infos and data stream states
            for data_stream in trace:
                info = _LttngLiveViewerStreamInfo(
                    stream_id,
                    trace_id,
                    False,
                    data_stream.path,
                    data_stream.channel_name,
                )
                self._stream_infos.append(info)
                self._ds_states[stream_id] = _LttngLiveViewerSessionDataStreamState(
                    self, info, data_stream
                )
                stream_id += 1

            # Metadata stream -> stream info and metadata stream state
            info = _LttngLiveViewerStreamInfo(
                stream_id, trace_id, True, trace.metadata_stream.path, 'metadata'
            )
            self._stream_infos.append(info)
            self._ms_states[stream_id] = _LttngLiveViewerSessionMetadataStreamState(
                self, info, trace.metadata_stream
            )
            stream_id += 1

        self._is_attached = False
        fmt = 'Built tracing session state: id={}, name="{}"'
        logging.info(fmt.format(tc_descr.info.tracing_session_id, tc_descr.info.name))

    @property
    def tracing_session_descriptor(self):
        return self._tc_descr

    @property
    def data_stream_states(self):
        return self._ds_states

    @property
    def metadata_stream_states(self):
        return self._ms_states

    @property
    def stream_infos(self):
        return self._stream_infos

    @property
    def has_new_metadata(self):
        return any([not ms.is_sent for ms in self._ms_states.values()])

    @property
    def is_attached(self):
        return self._is_attached

    @is_attached.setter
    def is_attached(self, value):
        self._is_attached = value


# An LTTng live viewer session manages a view on tracing sessions
# and replies to commands accordingly.
class _LttngLiveViewerSession:
    def __init__(
        self,
        viewer_session_id,
        tracing_session_descriptors,
        max_query_data_response_size,
    ):
        self._viewer_session_id = viewer_session_id
        self._ts_states = {}
        self._stream_states = {}
        self._max_query_data_response_size = max_query_data_response_size
        total_stream_infos = 0

        for ts_descr in tracing_session_descriptors:
            ts_state = _LttngLiveViewerSessionTracingSessionState(
                ts_descr, total_stream_infos
            )
            ts_id = ts_state.tracing_session_descriptor.info.tracing_session_id
            self._ts_states[ts_id] = ts_state
            total_stream_infos += len(ts_state.stream_infos)

            # Update session's stream states to have the new states
            self._stream_states.update(ts_state.data_stream_states)
            self._stream_states.update(ts_state.metadata_stream_states)

        self._command_handlers = {
            _LttngLiveViewerAttachToTracingSessionCommand: self._handle_attach_to_tracing_session_command,
            _LttngLiveViewerCreateViewerSessionCommand: self._handle_create_viewer_session_command,
            _LttngLiveViewerDetachFromTracingSessionCommand: self._handle_detach_from_tracing_session_command,
            _LttngLiveViewerGetDataStreamPacketDataCommand: self._handle_get_data_stream_packet_data_command,
            _LttngLiveViewerGetMetadataStreamDataCommand: self._handle_get_metadata_stream_data_command,
            _LttngLiveViewerGetNewStreamInfosCommand: self._handle_get_new_stream_infos_command,
            _LttngLiveViewerGetNextDataStreamIndexEntryCommand: self._handle_get_next_data_stream_index_entry_command,
            _LttngLiveViewerGetTracingSessionInfosCommand: self._handle_get_tracing_session_infos_command,
        }

    @property
    def viewer_session_id(self):
        return self._viewer_session_id

    def _get_tracing_session_state(self, tracing_session_id):
        if tracing_session_id not in self._ts_states:
            raise UnexpectedInput(
                'Unknown tracing session ID {}'.format(tracing_session_id)
            )

        return self._ts_states[tracing_session_id]

    def _get_stream_state(self, stream_id):
        if stream_id not in self._stream_states:
            UnexpectedInput('Unknown stream ID {}'.format(stream_id))

        return self._stream_states[stream_id]

    def handle_command(self, cmd):
        logging.info(
            'Handling command in viewer session: cmd-cls-name={}'.format(
                cmd.__class__.__name__
            )
        )
        cmd_type = type(cmd)

        if cmd_type not in self._command_handlers:
            raise UnexpectedInput(
                'Unexpected command: cmd-cls-name={}'.format(cmd.__class__.__name__)
            )

        return self._command_handlers[cmd_type](cmd)

    def _handle_attach_to_tracing_session_command(self, cmd):
        fmt = 'Handling "attach to tracing session" command: ts-id={}, offset={}, seek-type={}'
        logging.info(fmt.format(cmd.tracing_session_id, cmd.offset, cmd.seek_type))
        ts_state = self._get_tracing_session_state(cmd.tracing_session_id)
        info = ts_state.tracing_session_descriptor.info

        if ts_state.is_attached:
            raise UnexpectedInput(
                'Cannot attach to tracing session `{}`: viewer is already attached'.format(
                    info.name
                )
            )

        ts_state.is_attached = True
        status = _LttngLiveViewerAttachToTracingSessionReply.Status.OK
        return _LttngLiveViewerAttachToTracingSessionReply(
            status, ts_state.stream_infos
        )

    def _handle_detach_from_tracing_session_command(self, cmd):
        fmt = 'Handling "detach from tracing session" command: ts-id={}'
        logging.info(fmt.format(cmd.tracing_session_id))
        ts_state = self._get_tracing_session_state(cmd.tracing_session_id)
        info = ts_state.tracing_session_descriptor.info

        if not ts_state.is_attached:
            raise UnexpectedInput(
                'Cannot detach to tracing session `{}`: viewer is not attached'.format(
                    info.name
                )
            )

        ts_state.is_attached = False
        status = _LttngLiveViewerDetachFromTracingSessionReply.Status.OK
        return _LttngLiveViewerDetachFromTracingSessionReply(status)

    def _handle_get_next_data_stream_index_entry_command(self, cmd):
        fmt = 'Handling "get next data stream index entry" command: stream-id={}'
        logging.info(fmt.format(cmd.stream_id))
        stream_state = self._get_stream_state(cmd.stream_id)

        if type(stream_state) is not _LttngLiveViewerSessionDataStreamState:
            raise UnexpectedInput(
                'Stream with ID {} is not a data stream'.format(cmd.stream_id)
            )

        if stream_state.cur_index_entry is None:
            # The viewer is done reading this stream
            status = _LttngLiveViewerGetNextDataStreamIndexEntryReply.Status.HUP

            # Dummy data stream index entry to use with the `HUP` status
            # (the reply needs one, but the viewer ignores it)
            index_entry = _LttngDataStreamIndexEntry(0, 0, 0, 0, 0, 0, 0)

            return _LttngLiveViewerGetNextDataStreamIndexEntryReply(
                status, index_entry, False, False
            )

        # The viewer only checks the `has_new_metadata` flag if the
        # reply's status is `OK`, so we need to provide an index here
        has_new_metadata = stream_state.tracing_session_state.has_new_metadata
        status = _LttngLiveViewerGetNextDataStreamIndexEntryReply.Status.OK
        reply = _LttngLiveViewerGetNextDataStreamIndexEntryReply(
            status, stream_state.cur_index_entry, has_new_metadata, False
        )
        stream_state.goto_next_index_entry()
        return reply

    def _handle_get_data_stream_packet_data_command(self, cmd):
        fmt = 'Handling "get data stream packet data" command: stream-id={}, offset={}, req-length={}'
        logging.info(fmt.format(cmd.stream_id, cmd.offset, cmd.req_length))
        stream_state = self._get_stream_state(cmd.stream_id)
        data_response_length = cmd.req_length

        if type(stream_state) is not _LttngLiveViewerSessionDataStreamState:
            raise UnexpectedInput(
                'Stream with ID {} is not a data stream'.format(cmd.stream_id)
            )

        if stream_state.tracing_session_state.has_new_metadata:
            status = _LttngLiveViewerGetDataStreamPacketDataReply.Status.ERROR
            return _LttngLiveViewerGetDataStreamPacketDataReply(
                status, bytes(), True, False
            )

        if self._max_query_data_response_size:
            # Enforce a server side limit on the query requested length.
            # To ensure that the transaction terminate take the minimum of both
            # value.
            data_response_length = min(
                cmd.req_length, self._max_query_data_response_size
            )
            fmt = 'Limiting "get data stream packet data" command: req-length={} actual response size={}'
            logging.info(fmt.format(cmd.req_length, data_response_length))

        data = stream_state.data_stream.get_data(cmd.offset, data_response_length)
        status = _LttngLiveViewerGetDataStreamPacketDataReply.Status.OK
        return _LttngLiveViewerGetDataStreamPacketDataReply(status, data, False, False)

    def _handle_get_metadata_stream_data_command(self, cmd):
        fmt = 'Handling "get metadata stream data" command: stream-id={}'
        logging.info(fmt.format(cmd.stream_id))
        stream_state = self._get_stream_state(cmd.stream_id)

        if type(stream_state) is not _LttngLiveViewerSessionMetadataStreamState:
            raise UnexpectedInput(
                'Stream with ID {} is not a metadata stream'.format(cmd.stream_id)
            )

        if stream_state.is_sent:
            status = _LttngLiveViewerGetMetadataStreamDataContentReply.Status.NO_NEW
            return _LttngLiveViewerGetMetadataStreamDataContentReply(status, bytes())

        stream_state.is_sent = True
        status = _LttngLiveViewerGetMetadataStreamDataContentReply.Status.OK
        return _LttngLiveViewerGetMetadataStreamDataContentReply(
            status, stream_state.metadata_stream.data
        )

    def _handle_get_new_stream_infos_command(self, cmd):
        fmt = 'Handling "get new stream infos" command: ts-id={}'
        logging.info(fmt.format(cmd.tracing_session_id))

        # As of this version, all the tracing session's stream infos are
        # always given to the viewer when sending the "attach to tracing
        # session" reply, so there's nothing new here. Return the `HUP`
        # status as, if we're handling this command, the viewer consumed
        # all the existing data streams.
        status = _LttngLiveViewerGetNewStreamInfosReply.Status.HUP
        return _LttngLiveViewerGetNewStreamInfosReply(status, [])

    def _handle_get_tracing_session_infos_command(self, cmd):
        logging.info('Handling "get tracing session infos" command.')
        infos = [
            tss.tracing_session_descriptor.info for tss in self._ts_states.values()
        ]
        infos.sort(key=lambda info: info.name)
        return _LttngLiveViewerGetTracingSessionInfosReply(infos)

    def _handle_create_viewer_session_command(self, cmd):
        logging.info('Handling "create viewer session" command.')
        status = _LttngLiveViewerCreateViewerSessionReply.Status.OK

        # This does nothing here. In the LTTng relay daemon, it
        # allocates the viewer session's state.
        return _LttngLiveViewerCreateViewerSessionReply(status)


# An LTTng live TCP server.
#
# On creation, it binds to `localhost` with an OS-assigned TCP port. It writes
# the decimal TCP port number to a temporary port file.  It renames the
# temporary port file to `port_filename`.
#
# `tracing_session_descriptors` is a list of tracing session descriptors
# (`LttngTracingSessionDescriptor`) to serve.
#
# This server accepts a single viewer (client).
#
# When the viewer closes the connection, the server's constructor
# returns.
class LttngLiveServer:
    def __init__(
        self, port_filename, tracing_session_descriptors, max_query_data_response_size
    ):
        logging.info('Server configuration:')

        logging.info('  Port file name: `{}`'.format(port_filename))

        if max_query_data_response_size is not None:
            logging.info(
                '  Maximum response data query size: `{}`'.format(
                    max_query_data_response_size
                )
            )

        for ts_descr in tracing_session_descriptors:
            info = ts_descr.info
            fmt = '  TS descriptor: name="{}", id={}, hostname="{}", live-timer-freq={}, client-count={}, stream-count={}:'
            logging.info(
                fmt.format(
                    info.name,
                    info.tracing_session_id,
                    info.hostname,
                    info.live_timer_freq,
                    info.client_count,
                    info.stream_count,
                )
            )

            for trace in ts_descr.traces:
                logging.info('    Trace: path="{}"'.format(trace.path))

        self._ts_descriptors = tracing_session_descriptors
        self._max_query_data_response_size = max_query_data_response_size
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._codec = _LttngLiveViewerProtocolCodec()

        # Port 0: OS assigns an unused port
        serv_addr = ('localhost', 0)
        self._sock.bind(serv_addr)
        self._write_port_to_file(port_filename)

        try:
            self._listen()
        finally:
            self._sock.close()
            logging.info('Closed connection and socket.')

    @property
    def _server_port(self):
        return self._sock.getsockname()[1]

    def _recv_command(self):
        data = bytes()

        while True:
            logging.info('Waiting for viewer command.')
            buf = self._conn.recv(128)

            if not buf:
                logging.info('Client closed connection.')

                if data:
                    raise UnexpectedInput(
                        'Client closed connection after having sent {} command bytes.'.format(
                            len(data)
                        )
                    )

                return

            logging.info('Received data from viewer: length={}'.format(len(buf)))

            data += buf

            try:
                cmd = self._codec.decode(data)
            except struct.error as exc:
                raise UnexpectedInput('Malformed command: {}'.format(exc)) from exc

            if cmd is not None:
                logging.info(
                    'Received command from viewer: cmd-cls-name={}'.format(
                        cmd.__class__.__name__
                    )
                )
                return cmd

    def _send_reply(self, reply):
        data = self._codec.encode(reply)
        logging.info(
            'Sending reply to viewer: reply-cls-name={}, length={}'.format(
                reply.__class__.__name__, len(data)
            )
        )
        self._conn.sendall(data)

    def _handle_connection(self):
        # First command must be "connect"
        cmd = self._recv_command()

        if type(cmd) is not _LttngLiveViewerConnectCommand:
            raise UnexpectedInput(
                'First command is not "connect": cmd-cls-name={}'.format(
                    cmd.__class__.__name__
                )
            )

        # Create viewer session (arbitrary ID 23)
        logging.info(
            'LTTng live viewer connected: version={}.{}'.format(cmd.major, cmd.minor)
        )
        viewer_session = _LttngLiveViewerSession(
            23, self._ts_descriptors, self._max_query_data_response_size
        )

        # Send "connect" reply
        self._send_reply(
            _LttngLiveViewerConnectReply(viewer_session.viewer_session_id, 2, 10)
        )

        # Make the viewer session handle the remaining commands
        while True:
            cmd = self._recv_command()

            if cmd is None:
                # Connection closed (at an expected location within the
                # conversation)
                return

            self._send_reply(viewer_session.handle_command(cmd))

    def _listen(self):
        logging.info('Listening: port={}'.format(self._server_port))
        # Backlog must be present for Python version < 3.5.
        # 128 is an arbitrary number since we expect only 1 connection anyway.
        self._sock.listen(128)
        self._conn, viewer_addr = self._sock.accept()
        logging.info(
            'Accepted viewer: addr={}:{}'.format(viewer_addr[0], viewer_addr[1])
        )

        try:
            self._handle_connection()
        finally:
            self._conn.close()

    def _write_port_to_file(self, port_filename):
        # Write the port number to a temporary file.
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as tmp_port_file:
            print(self._server_port, end='', file=tmp_port_file)

        # Rename temporary file to real file
        os.replace(tmp_port_file.name, port_filename)
        logging.info(
            'Renamed port file: src-path="{}", dst-path="{}"'.format(
                tmp_port_file.name, port_filename
            )
        )


# A tracing session descriptor.
#
# In the constructor, `traces` is a list of LTTng traces (`LttngTrace`
# objects).
class LttngTracingSessionDescriptor:
    def __init__(
        self, name, tracing_session_id, hostname, live_timer_freq, client_count, traces
    ):
        for trace in traces:
            if name not in trace.path:
                fmt = 'Tracing session name must be part of every trace path (`{}` not found in `{}`)'
                raise ValueError(fmt.format(name, trace.path))

        self._traces = traces
        stream_count = sum([len(t) + 1 for t in traces])
        self._info = _LttngLiveViewerTracingSessionInfo(
            tracing_session_id,
            live_timer_freq,
            client_count,
            stream_count,
            hostname,
            name,
        )

    @property
    def traces(self):
        return self._traces

    @property
    def info(self):
        return self._info


def _tracing_session_descriptors_from_arg(string):
    # Format is:
    #     NAME,ID,HOSTNAME,FREQ,CLIENTS,TRACEPATH[,TRACEPATH]...
    parts = string.split(',')
    name = parts[0]
    tracing_session_id = int(parts[1])
    hostname = parts[2]
    live_timer_freq = int(parts[3])
    client_count = int(parts[4])
    traces = [LttngTrace(path) for path in parts[5:]]
    return LttngTracingSessionDescriptor(
        name, tracing_session_id, hostname, live_timer_freq, client_count, traces
    )


def _loglevel_parser(string):
    loglevels = {'info': logging.INFO, 'warning': logging.WARNING}
    if string not in loglevels:
        msg = "{} is not a valid loglevel".format(string)
        raise argparse.ArgumentTypeError(msg)
    return loglevels[string]


if __name__ == '__main__':
    logging.basicConfig(format='# %(asctime)-25s%(message)s')
    parser = argparse.ArgumentParser(
        description='LTTng-live protocol mocker', add_help=False
    )
    parser.add_argument(
        '--log-level',
        default='warning',
        choices=['info', 'warning'],
        help='The loglevel to be used.',
    )

    loglevel_namespace, remaining_args = parser.parse_known_args()
    logging.getLogger().setLevel(_loglevel_parser(loglevel_namespace.log_level))

    parser.add_argument(
        '--port-filename',
        help='The final port file. This file is present when the server is ready to receive connection.',
        required=True,
    )
    parser.add_argument(
        '--max-query-data-response-size',
        type=int,
        help='The maximum size of control data response in bytes',
    )
    parser.add_argument(
        'sessions',
        nargs="+",
        metavar="SESSION",
        type=_tracing_session_descriptors_from_arg,
        help='A session configuration. There is no space after comma. Format is: NAME,ID,HOSTNAME,FREQ,CLIENTS,TRACEPATH[,TRACEPATH]....',
    )
    parser.add_argument(
        '-h',
        '--help',
        action='help',
        default=argparse.SUPPRESS,
        help='Show this help message and exit.',
    )

    args = parser.parse_args(args=remaining_args)
    try:
        LttngLiveServer(
            args.port_filename, args.sessions, args.max_query_data_response_size
        )
    except UnexpectedInput as exc:
        logging.error(str(exc))
        print(exc, file=sys.stderr)
        sys.exit(1)



# Copyright (c) 2016, Matt Layman


class Line(object):
    """Base type for TAP data.

    TAP is a line based protocol. Thus, the most primitive type is a line.
    """
    @property
    def category(self):
        raise NotImplementedError


class Result(Line):
    """Information about an individual test line."""

    def __init__(
            self, ok, number=None, description='', directive=None,
            diagnostics=None):
        self._ok = ok
        if number:
            self._number = int(number)
        else:
            # The number may be an empty string so explicitly set to None.
            self._number = None
        self._description = description
        self.directive = directive
        self.diagnostics = diagnostics

    @property
    def category(self):
        """:returns: ``test``"""
        return 'test'

    @property
    def ok(self):
        """Get the ok status.

        :rtype: bool
        """
        return self._ok

    @property
    def number(self):
        """Get the test number.

        :rtype: int
        """
        return self._number

    @property
    def description(self):
        """Get the description."""
        return self._description

    @property
    def skip(self):
        """Check if this test was skipped.

        :rtype: bool
        """
        return self.directive.skip

    @property
    def todo(self):
        """Check if this test was a TODO.

        :rtype: bool
        """
        return self.directive.todo

    def __str__(self):
        is_not = ''
        if not self.ok:
            is_not = 'not '
        directive = ''
        if self.directive is not None and self.directive.text:
            directive = ' # {0}'.format(self.directive.text)
        diagnostics = ''
        if self.diagnostics is not None:
            diagnostics = '\n' + self.diagnostics.rstrip()
        return "{0}ok {1} - {2}{3}{4}".format(
            is_not, self.number, self.description, directive, diagnostics)


class Plan(Line):
    """A plan line to indicate how many tests to expect."""

    def __init__(self, expected_tests, directive=None):
        self._expected_tests = expected_tests
        self.directive = directive

    @property
    def category(self):
        """:returns: ``plan``"""
        return 'plan'

    @property
    def expected_tests(self):
        """Get the number of expected tests.

        :rtype: int
        """
        return self._expected_tests

    @property
    def skip(self):
        """Check if this plan should skip the file.

        :rtype: bool
        """
        return self.directive.skip


class Diagnostic(Line):
    """A diagnostic line (i.e. anything starting with a hash)."""

    def __init__(self, text):
        self._text = text

    @property
    def category(self):
        """:returns: ``diagnostic``"""
        return 'diagnostic'

    @property
    def text(self):
        """Get the text."""
        return self._text


class Bail(Line):
    """A bail out line (i.e. anything starting with 'Bail out!')."""

    def __init__(self, reason):
        self._reason = reason

    @property
    def category(self):
        """:returns: ``bail``"""
        return 'bail'

    @property
    def reason(self):
        """Get the reason."""
        return self._reason


class Version(Line):
    """A version line (i.e. of the form 'TAP version 13')."""

    def __init__(self, version):
        self._version = version

    @property
    def category(self):
        """:returns: ``version``"""
        return 'version'

    @property
    def version(self):
        """Get the version number.

        :rtype: int
        """
        return self._version


class Unknown(Line):
    """A line that represents something that is not a known TAP line.

    This exists for the purpose of a Null Object pattern.
    """
    @property
    def category(self):
        """:returns: ``unknown``"""
        return 'unknown'



#
# Copyright (C) 2019 EfficiOS Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; only version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

from bt2 import native_bt
import bt2
import unittest


class FailingIter(bt2._UserMessageIterator):
    def __next__(self):
        raise ValueError('User message iterator is failing')


class SourceWithFailingIter(
    bt2._UserSourceComponent, message_iterator_class=FailingIter
):
    def __init__(self, config, params, obj):
        self._add_output_port('out')


class SourceWithFailingInit(
    bt2._UserSourceComponent, message_iterator_class=FailingIter
):
    def __init__(self, config, params, obj):
        raise ValueError('Source is failing')


class WorkingSink(bt2._UserSinkComponent):
    def __init__(self, config, params, obj):
        self._in = self._add_input_port('in')

    def _user_graph_is_configured(self):
        self._iter = self._create_message_iterator(self._in)

    def _user_consume(self):
        next(self._iter)


class SinkWithExceptionChaining(bt2._UserSinkComponent):
    def __init__(self, config, params, obj):
        self._in = self._add_input_port('in')

    def _user_graph_is_configured(self):
        self._iter = self._create_message_iterator(self._in)

    def _user_consume(self):
        try:
            next(self._iter)
        except bt2._Error as e:
            raise ValueError('oops') from e


class SinkWithFailingQuery(bt2._UserSinkComponent):
    def _user_consume(self):
        pass

    @staticmethod
    def _user_query(priv_executor, obj, params, method_obj):
        raise ValueError('Query is failing')


class ErrorTestCase(unittest.TestCase):
    def _run_failing_graph(self, source_cc, sink_cc):
        with self.assertRaises(bt2._Error) as ctx:
            graph = bt2.Graph()
            src = graph.add_component(source_cc, 'src')
            snk = graph.add_component(sink_cc, 'snk')
            graph.connect_ports(src.output_ports['out'], snk.input_ports['in'])
            graph.run()

        return ctx.exception

    def test_current_thread_error_none(self):
        # When a bt2._Error is raised, it steals the current thread's error.
        # Verify that it is now NULL.
        self._run_failing_graph(SourceWithFailingInit, WorkingSink)
        self.assertIsNone(native_bt.current_thread_take_error())

    def test_len(self):
        exc = self._run_failing_graph(SourceWithFailingIter, WorkingSink)

        # The exact number of causes is not too important (it can change if we
        # append more or less causes along the way), but the idea is to verify is
        # has a value that makes sense.
        self.assertEqual(len(exc), 4)

    def test_iter(self):
        exc = self._run_failing_graph(SourceWithFailingIter, WorkingSink)

        for c in exc:
            # Each cause is an instance of _ErrorCause (including subclasses).
            self.assertIsInstance(c, bt2._ErrorCause)

    def test_getitem(self):
        exc = self._run_failing_graph(SourceWithFailingIter, WorkingSink)

        for i in range(len(exc)):
            c = exc[i]
            # Each cause is an instance of _ErrorCause (including subclasses).
            self.assertIsInstance(c, bt2._ErrorCause)

    def test_getitem_indexerror(self):
        exc = self._run_failing_graph(SourceWithFailingIter, WorkingSink)

        with self.assertRaises(IndexError):
            exc[len(exc)]

    def test_exception_chaining(self):
        # Test that if we do:
        #
        # try:
        #     ...
        # except bt2._Error as exc:
        #     raise ValueError('oh noes') from exc
        #
        # We are able to fetch the causes of the original bt2._Error in the
        # exception chain.  Also, each exception in the chain should become one
        # cause once caught.
        exc = self._run_failing_graph(SourceWithFailingIter, SinkWithExceptionChaining)

        self.assertEqual(len(exc), 5)

        self.assertIsInstance(exc[0], bt2._MessageIteratorErrorCause)
        self.assertEqual(exc[0].component_class_name, 'SourceWithFailingIter')
        self.assertIn('ValueError: User message iterator is failing', exc[0].message)

        self.assertIsInstance(exc[1], bt2._ErrorCause)

        self.assertIsInstance(exc[2], bt2._ComponentErrorCause)
        self.assertEqual(exc[2].component_class_name, 'SinkWithExceptionChaining')
        self.assertIn(
            'unexpected error: cannot advance the message iterator', exc[2].message
        )

        self.assertIsInstance(exc[3], bt2._ComponentErrorCause)
        self.assertEqual(exc[3].component_class_name, 'SinkWithExceptionChaining')
        self.assertIn('ValueError: oops', exc[3].message)

        self.assertIsInstance(exc[4], bt2._ErrorCause)

    def _common_cause_tests(self, cause):
        self.assertIsInstance(cause.module_name, str)
        self.assertIsInstance(cause.file_name, str)
        self.assertIsInstance(cause.line_number, int)

    def test_unknown_error_cause(self):
        exc = self._run_failing_graph(SourceWithFailingIter, SinkWithExceptionChaining)
        cause = exc[-1]
        self.assertIs(type(cause), bt2._ErrorCause)
        self._common_cause_tests(cause)

    def test_component_error_cause(self):
        exc = self._run_failing_graph(SourceWithFailingInit, SinkWithExceptionChaining)
        cause = exc[0]
        self.assertIs(type(cause), bt2._ComponentErrorCause)
        self._common_cause_tests(cause)

        self.assertIn('Source is failing', cause.message)
        self.assertEqual(cause.component_name, 'src')
        self.assertEqual(cause.component_class_type, bt2.ComponentClassType.SOURCE)
        self.assertEqual(cause.component_class_name, 'SourceWithFailingInit')
        self.assertIsNone(cause.plugin_name)

    def test_component_class_error_cause(self):
        q = bt2.QueryExecutor(SinkWithFailingQuery, 'hello')

        with self.assertRaises(bt2._Error) as ctx:
            q.query()

        cause = ctx.exception[0]
        self.assertIs(type(cause), bt2._ComponentClassErrorCause)
        self._common_cause_tests(cause)

        self.assertIn('Query is failing', cause.message)

        self.assertEqual(cause.component_class_type, bt2.ComponentClassType.SINK)
        self.assertEqual(cause.component_class_name, 'SinkWithFailingQuery')
        self.assertIsNone(cause.plugin_name)

    def test_message_iterator_error_cause(self):
        exc = self._run_failing_graph(SourceWithFailingIter, SinkWithExceptionChaining)
        cause = exc[0]
        self.assertIs(type(cause), bt2._MessageIteratorErrorCause)
        self._common_cause_tests(cause)

        self.assertIn('User message iterator is failing', cause.message)
        self.assertEqual(cause.component_name, 'src')
        self.assertEqual(cause.component_output_port_name, 'out')
        self.assertEqual(cause.component_class_type, bt2.ComponentClassType.SOURCE)
        self.assertEqual(cause.component_class_name, 'SourceWithFailingIter')
        self.assertIsNone(cause.plugin_name)

    def test_str(self):
        # Test __str__.  We don't need to test the precise format used, but
        # just that it doesn't miserably crash and that it contains some
        # expected bits.
        exc = self._run_failing_graph(SourceWithFailingIter, SinkWithExceptionChaining)
        s = str(exc)
        self.assertIn("[src (out): 'source.SourceWithFailingIter']", s)
        self.assertIn('ValueError: oops', s)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (C) 2019 EfficiOS Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; only version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import unittest
import bt2
from bt2 import component as bt2_component


class UserComponentTestCase(unittest.TestCase):
    @staticmethod
    def _create_comp(comp_cls, name=None, log_level=bt2.LoggingLevel.NONE):
        graph = bt2.Graph()

        if name is None:
            name = 'comp'

        return graph.add_component(comp_cls, name, logging_level=log_level)

    def test_name(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self.name, 'yaes')

            def _user_consume(self):
                pass

        self._create_comp(MySink, 'yaes')

    def test_logging_level(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self.logging_level, bt2.LoggingLevel.INFO)

            def _user_consume(self):
                pass

        self._create_comp(MySink, 'yaes', bt2.LoggingLevel.INFO)

    def test_graph_mip_version(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self._graph_mip_version, 0)

            def _user_consume(self):
                pass

        self._create_comp(MySink, 'yaes', bt2.LoggingLevel.INFO)

    def test_class(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertEqual(comp_self.cls, MySink)

            def _user_consume(self):
                pass

        self._create_comp(MySink)

    def test_addr(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                self.assertIsInstance(comp_self.addr, int)
                self.assertNotEqual(comp_self.addr, 0)

            def _user_consume(self):
                pass

        self._create_comp(MySink)

    def test_finalize(self):
        finalized = False

        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            def _user_finalize(comp_self):
                nonlocal finalized
                finalized = True

        graph = bt2.Graph()
        comp = graph.add_component(MySink, 'lel')

        del graph
        del comp
        self.assertTrue(finalized)

    def test_source_component_config(self):
        class MySource(
            bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            def __init__(comp_self, config, params, obj):
                nonlocal cfg_type
                cfg_type = type(config)

        cfg_type = None
        self._create_comp(MySource)
        self.assertIs(cfg_type, bt2_component._UserSourceComponentConfiguration)

    def test_filter_component_config(self):
        class MyFilter(
            bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            def __init__(comp_self, config, params, obj):
                nonlocal cfg_type
                cfg_type = type(config)

        cfg_type = None
        self._create_comp(MyFilter)
        self.assertIs(cfg_type, bt2_component._UserFilterComponentConfiguration)

    def test_sink_component_config(self):
        class MySink(bt2._UserSinkComponent):
            def __init__(comp_self, config, params, obj):
                nonlocal cfg_type
                cfg_type = type(config)

            def _user_consume(self):
                pass

        cfg_type = None
        self._create_comp(MySink)
        self.assertIs(cfg_type, bt2_component._UserSinkComponentConfiguration)


class GenericComponentTestCase(unittest.TestCase):
    @staticmethod
    def _create_comp(comp_cls, name=None, log_level=bt2.LoggingLevel.NONE):
        graph = bt2.Graph()

        if name is None:
            name = 'comp'

        return graph.add_component(comp_cls, name, logging_level=log_level)

    def test_name(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink, 'yaes')
        self.assertEqual(comp.name, 'yaes')

    def test_logging_level(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink, 'yaes', bt2.LoggingLevel.WARNING)
        self.assertEqual(comp.logging_level, bt2.LoggingLevel.WARNING)

    def test_class(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink)
        self.assertEqual(comp.cls, MySink)

    def test_addr(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        comp = self._create_comp(MySink)
        self.assertIsInstance(comp.addr, int)
        self.assertNotEqual(comp.addr, 0)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (C) 2019 EfficiOS Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; only version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import bt2


@bt2.plugin_component_class
class ThisIsASource(
    bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
):
    pass


@bt2.plugin_component_class
class ThisIsAFilter(
    bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
):
    pass


@bt2.plugin_component_class
class ThisIsASink(bt2._UserSinkComponent):
    def _user_consume(self):
        pass


bt2.register_plugin(
    __name__,
    "this-is-a-plugin",
    version=(1, 2, 3, 'bob'),
    description='A plugin',
    author='Jorge Mario Bergoglio',
    license='The license',
)



#
# Copyright (C) 2019 EfficiOS Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; only version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import bt2
import unittest

from utils import get_default_trace_class, create_const_field


def get_const_signed_integer_range(int_ranges):
    def range_setter(field):
        field.value = 12

    tc = get_default_trace_class()
    fc = tc.create_signed_enumeration_field_class(32)
    fc.add_mapping('something', bt2.SignedIntegerRangeSet(int_ranges))

    return create_const_field(tc, fc, range_setter).cls['something'].ranges


def get_const_unsigned_integer_range(int_ranges):
    def range_setter(field):
        field.value = 12

    tc = get_default_trace_class()
    fc = tc.create_unsigned_enumeration_field_class(32)
    fc.add_mapping('something', bt2.UnsignedIntegerRangeSet(int_ranges))

    return create_const_field(tc, fc, range_setter).cls['something'].ranges


class _IntegerRangeTestCase:
    def setUp(self):
        self._rg = self._CLS(self._def_lower, self._def_upper)
        self._const_rg = list(
            self._GET_CONST_RANGE_SET([(self._def_lower, self._def_upper)])
        )[0]

    def test_create(self):
        self.assertEqual(self._rg.lower, self._def_lower)
        self.assertEqual(self._rg.upper, self._def_upper)
        self.assertIs(type(self._rg), self._CLS)

    def test_const_create(self):
        self.assertEqual(self._const_rg.lower, self._def_lower)
        self.assertEqual(self._const_rg.upper, self._def_upper)
        self.assertIs(type(self._const_rg), self._CONST_CLS)

    def test_create_same(self):
        rg = self._CLS(self._def_lower, self._def_lower)
        self.assertEqual(rg.lower, self._def_lower)
        self.assertEqual(rg.upper, self._def_lower)

    def test_create_single(self):
        rg = self._CLS(self._def_lower)
        self.assertEqual(rg.lower, self._def_lower)
        self.assertEqual(rg.upper, self._def_lower)

    def test_create_wrong_type_lower(self):
        with self.assertRaises(TypeError):
            self._CLS(19.3, self._def_upper)

    def test_create_wrong_type_upper(self):
        with self.assertRaises(TypeError):
            self._CLS(self._def_lower, 19.3)

    def test_create_out_of_bound_lower(self):
        with self.assertRaises(ValueError):
            self._CLS(self._oob_lower, self._def_upper)

    def test_create_out_of_bound_upper(self):
        with self.assertRaises(ValueError):
            self._CLS(self._def_lower, self._oob_upper)

    def test_create_lower_gt_upper(self):
        with self.assertRaises(ValueError):
            self._CLS(self._def_lower, self._def_lower - 1)

    def test_contains_lower(self):
        self.assertTrue(self._rg.contains(self._def_lower))

    def test_contains_upper(self):
        self.assertTrue(self._rg.contains(self._def_upper))

    def test_contains_avg(self):
        avg = (self._def_lower + self._def_upper) // 2
        self.assertTrue(self._rg.contains(avg))

    def test_contains_wrong_type(self):
        with self.assertRaises(TypeError):
            self._rg.contains('allo')

    def test_contains_out_of_bound(self):
        with self.assertRaises(ValueError):
            self._rg.contains(self._oob_upper)

    def test_eq(self):
        rg = self._CLS(self._def_lower, self._def_upper)
        self.assertEqual(rg, self._rg)

    def test_const_eq(self):
        rg = list(self._GET_CONST_RANGE_SET([(self._def_lower, self._def_upper)]))[0]
        self.assertEqual(rg, self._const_rg)

    def test_const_nonconst_eq(self):
        self.assertEqual(self._rg, self._const_rg)

    def test_ne(self):
        rg = self._CLS(self._def_lower, self._def_upper - 1)
        self.assertNotEqual(rg, self._rg)

    def test_const_ne(self):
        rg = list(self._GET_CONST_RANGE_SET([(self._def_lower, self._def_upper - 1)]))[
            0
        ]
        self.assertNotEqual(rg, self._const_rg)

    def test_ne_other_type(self):
        self.assertNotEqual(self._rg, 48)


class UnsignedIntegerRangeTestCase(_IntegerRangeTestCase, unittest.TestCase):
    _CLS = bt2.UnsignedIntegerRange
    _CONST_CLS = bt2._UnsignedIntegerRangeConst
    _GET_CONST_RANGE_SET = staticmethod(get_const_unsigned_integer_range)
    _def_lower = 23
    _def_upper = 18293
    _oob_lower = -1
    _oob_upper = 1 << 64


class SignedIntegerRangeTestCase(_IntegerRangeTestCase, unittest.TestCase):
    _CLS = bt2.SignedIntegerRange
    _CONST_CLS = bt2._SignedIntegerRangeConst
    _GET_CONST_RANGE_SET = staticmethod(get_const_signed_integer_range)
    _def_lower = -184
    _def_upper = 11547
    _oob_lower = -(1 << 63) - 1
    _oob_upper = 1 << 63


class _IntegerRangeSetTestCase:
    def setUp(self):
        self._rs = self._CLS((self._range1, self._range2, self._range3))
        self._const_rs = self._GET_CONST_RANGE_SET(
            [self._range1, self._range2, self._range3]
        )

    def test_create(self):
        self.assertEqual(len(self._rs), 3)
        self.assertIn(self._range1, self._rs)
        self.assertIn(self._range2, self._rs)
        self.assertIn(self._range3, self._rs)
        self.assertIs(type(self._range1), self._RANGE_CLS)

    def test_const_create(self):
        self.assertEqual(len(self._const_rs), 3)
        self.assertIn(self._range1, self._const_rs)
        self.assertIn(self._range2, self._const_rs)
        self.assertIn(self._range3, self._const_rs)
        self.assertIs(type(self._range1), self._RANGE_CLS)

    def test_create_tuples(self):
        rs = self._CLS(
            (
                (self._range1.lower, self._range1.upper),
                (self._range2.lower, self._range2.upper),
                (self._range3.lower, self._range3.upper),
            )
        )
        self.assertEqual(len(rs), 3)
        self.assertIn(self._range1, rs)
        self.assertIn(self._range2, rs)
        self.assertIn(self._range3, rs)

    def test_create_single(self):
        rs = self._CLS((self._range_same.lower,))
        self.assertEqual(len(rs), 1)
        self.assertIn(self._range_same, rs)

    def test_create_non_iter(self):
        with self.assertRaises(TypeError):
            self._rs = self._CLS(23)

    def test_create_wrong_elem_type(self):
        with self.assertRaises(TypeError):
            self._rs = self._CLS((self._range1, self._range2, 'lel'))

    def test_len(self):
        self.assertEqual(len(self._rs), 3)

    def test_contains(self):
        rs = self._CLS((self._range1, self._range2))
        self.assertIn(self._range1, rs)
        self.assertIn(self._range2, rs)
        self.assertNotIn(self._range3, rs)

    def test_contains_value(self):
        rs = self._CLS((self._range1, self._range2))
        self.assertTrue(rs.contains_value(self._range1.upper))
        self.assertTrue(rs.contains_value(self._range2.lower))
        self.assertFalse(rs.contains_value(self._range3.upper))

    def test_contains_value_wrong_type(self):
        with self.assertRaises(TypeError):
            self._rs.contains_value('meow')

    def test_iter(self):
        range_list = list(self._rs)
        self.assertIn(self._range1, range_list)
        self.assertIn(self._range2, range_list)
        self.assertIn(self._range3, range_list)

        for rg in range_list:
            self.assertIs(type(rg), self._RANGE_CLS)

    def test_const_iter(self):
        range_list = list(self._const_rs)
        self.assertIn(self._range1, range_list)
        self.assertIn(self._range2, range_list)
        self.assertIn(self._range3, range_list)

        for rg in range_list:
            self.assertIs(type(rg), self._CONST_RANGE_CLS)

    def test_empty(self):
        rs = self._CLS()
        self.assertEqual(len(rs), 0)
        self.assertEqual(len(list(rs)), 0)

    def test_add_range_obj(self):
        rs = self._CLS((self._range1,))
        self.assertEqual(len(rs), 1)
        rs.add(self._range2)
        self.assertEqual(len(rs), 2)
        self.assertIn(self._range2, rs)

    def test_const_add_range_obj(self):
        with self.assertRaises(AttributeError):
            self._const_rs.add((12, 4434))

    def test_discard_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self._rs.discard(self._range2)

    def test_eq_same_order(self):
        rs = self._CLS((self._range1, self._range2, self._range3))
        self.assertEqual(self._rs, rs)

    def test_eq_diff_order(self):
        rs = self._CLS((self._range1, self._range3, self._range2))
        self.assertEqual(self._rs, rs)

    def test_eq_same_addr(self):
        self.assertEqual(self._rs, self._rs)

    def test_ne_diff_len(self):
        rs = self._CLS((self._range1, self._range2))
        self.assertNotEqual(self._rs, rs)

    def test_ne_diff_values(self):
        rs1 = self._CLS((self._range1, self._range2))
        rs2 = self._CLS((self._range1, self._range3))
        self.assertNotEqual(rs1, rs2)

    def test_ne_incompatible_type(self):
        self.assertNotEqual(self._rs, object())


class UnsignedIntegerRangeSetTestCase(_IntegerRangeSetTestCase, unittest.TestCase):
    _CLS = bt2.UnsignedIntegerRangeSet
    _CONST_CLS = bt2._UnsignedIntegerRangeSetConst
    _RANGE_CLS = bt2.UnsignedIntegerRange
    _CONST_RANGE_CLS = bt2._UnsignedIntegerRangeConst
    _GET_CONST_RANGE_SET = staticmethod(get_const_unsigned_integer_range)

    def setUp(self):
        self._range1 = bt2.UnsignedIntegerRange(4, 192)
        self._range2 = bt2.UnsignedIntegerRange(17, 228)
        self._range3 = bt2.UnsignedIntegerRange(1000, 2000)
        self._range_same = bt2.UnsignedIntegerRange(1300, 1300)
        super().setUp()


class SignedIntegerRangeSetTestCase(_IntegerRangeSetTestCase, unittest.TestCase):
    _CLS = bt2.SignedIntegerRangeSet
    _CONST_CLS = bt2._SignedIntegerRangeSetConst
    _RANGE_CLS = bt2.SignedIntegerRange
    _CONST_RANGE_CLS = bt2._SignedIntegerRangeConst
    _GET_CONST_RANGE_SET = staticmethod(get_const_signed_integer_range)

    def setUp(self):
        self._range1 = bt2.SignedIntegerRange(-1484, -17)
        self._range2 = bt2.SignedIntegerRange(-101, 1500)
        self._range3 = bt2.SignedIntegerRange(1948, 2019)
        self._range_same = bt2.SignedIntegerRange(-1300, -1300)
        super().setUp()


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016, Matt Layman

import re


class Directive(object):
    """A representation of a result line directive."""

    skip_pattern = re.compile(
        r"""^SKIP\S*
            (?P<whitespace>\s*) # Optional whitespace.
            (?P<reason>.*)      # Slurp up the rest.""",
        re.IGNORECASE | re.VERBOSE)
    todo_pattern = re.compile(
        r"""^TODO\b             # The directive name
            (?P<whitespace>\s*) # Immediately following must be whitespace.
            (?P<reason>.*)      # Slurp up the rest.""",
        re.IGNORECASE | re.VERBOSE)

    def __init__(self, text):
        """Initialize the directive by parsing the text.

        The text is assumed to be everything after a '#\s*' on a result line.
        """
        self._text = text
        self._skip = False
        self._todo = False
        self._reason = None

        match = self.skip_pattern.match(text)
        if match:
            self._skip = True
            self._reason = match.group('reason')

        match = self.todo_pattern.match(text)
        if match:
            if match.group('whitespace'):
                self._todo = True
            else:
                # Catch the case where the directive has no descriptive text.
                if match.group('reason') == '':
                    self._todo = True
            self._reason = match.group('reason')

    @property
    def text(self):
        """Get the entire text."""
        return self._text

    @property
    def skip(self):
        """Check if the directive is a SKIP type."""
        return self._skip

    @property
    def todo(self):
        """Check if the directive is a TODO type."""
        return self._todo

    @property
    def reason(self):
        """Get the reason for the directive."""
        return self._reason



#
# Copyright (C) 2019 EfficiOS Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; only version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import unittest
import bt2


class UserComponentClassTestCase(unittest.TestCase):
    def _test_no_init(self, cls):
        with self.assertRaises(RuntimeError):
            cls()

    def test_no_init_source(self):
        class MySource(
            bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            pass

        self._test_no_init(MySource)

    def test_no_init_filter(self):
        class MyFilter(
            bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            pass

        self._test_no_init(MyFilter)

    def test_no_init_sink(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        self._test_no_init(MySink)

    def test_incomplete_source_no_msg_iter_cls(self):
        with self.assertRaises(bt2._IncompleteUserClass):

            class MySource(bt2._UserSourceComponent):
                pass

    def test_incomplete_source_wrong_msg_iter_cls_type(self):
        with self.assertRaises(bt2._IncompleteUserClass):

            class MySource(bt2._UserSourceComponent, message_iterator_class=int):
                pass

    def test_incomplete_filter_no_msg_iter_cls(self):
        with self.assertRaises(bt2._IncompleteUserClass):

            class MyFilter(bt2._UserFilterComponent):
                pass

    def test_incomplete_sink_no_consume_method(self):
        with self.assertRaises(bt2._IncompleteUserClass):

            class MySink(bt2._UserSinkComponent):
                pass

    def test_minimal_source(self):
        class MySource(
            bt2._UserSourceComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            pass

    def test_minimal_filter(self):
        class MyFilter(
            bt2._UserFilterComponent, message_iterator_class=bt2._UserMessageIterator
        ):
            pass

    def test_minimal_sink(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

    def test_default_name(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        self.assertEqual(MySink.name, 'MySink')

    def test_custom_name(self):
        class MySink(bt2._UserSinkComponent, name='salut'):
            def _user_consume(self):
                pass

        self.assertEqual(MySink.name, 'salut')

    def test_invalid_custom_name(self):
        with self.assertRaises(TypeError):

            class MySink(bt2._UserSinkComponent, name=23):
                def _user_consume(self):
                    pass

    def test_description(self):
        class MySink(bt2._UserSinkComponent):
            """
            The description.

            Bacon ipsum dolor amet ribeye t-bone corned beef, beef jerky
            porchetta burgdoggen prosciutto chicken frankfurter boudin
            hamburger doner bacon turducken. Sirloin shank sausage,
            boudin meatloaf alcatra meatball t-bone tongue pastrami
            cupim flank tenderloin.
            """

            def _user_consume(self):
                pass

        self.assertEqual(MySink.description, 'The description.')

    def test_empty_description_no_lines(self):
        class MySink(bt2._UserSinkComponent):
            # fmt: off
            """"""
            # fmt: on

            def _user_consume(self):
                pass

        self.assertIsNone(MySink.description)

    def test_empty_description_no_contents(self):
        class MySink(bt2._UserSinkComponent):
            # fmt: off
            """
            """
            # fmt: on

            def _user_consume(self):
                pass

        self.assertIsNone(MySink.description)

    def test_empty_description_single_line(self):
        class MySink(bt2._UserSinkComponent):
            """my description"""

            def _user_consume(self):
                pass

        self.assertEqual(MySink.description, "my description")

    def test_help(self):
        class MySink(bt2._UserSinkComponent):
            """
            The description.

            The help
            text is
            here.
            """

            def _user_consume(self):
                pass

        self.assertEqual(MySink.help, 'The help\ntext is\nhere.')

    def test_addr(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        self.assertIsInstance(MySink.addr, int)
        self.assertNotEqual(MySink.addr, 0)

    def test_query_not_implemented(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        with self.assertRaises(bt2.UnknownObject):
            bt2.QueryExecutor(MySink, 'obj', 23).query()

    def test_query_raises(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params):
                raise ValueError

        with self.assertRaises(bt2._Error):
            bt2.QueryExecutor(MySink, 'obj', 23).query()

    def test_query_wrong_return_type(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params, method_obj):
                return ...

        with self.assertRaises(bt2._Error):
            bt2.QueryExecutor(MySink, 'obj', 23).query()

    def test_query_params_none(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params, method_obj):
                nonlocal query_params
                query_params = params
                return None

        query_params = None
        params = None
        res = bt2.QueryExecutor(MySink, 'obj', params).query()
        self.assertEqual(query_params, params)
        self.assertIsNone(res)
        del query_params

    def test_query_logging_level(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params, method_obj):
                nonlocal query_log_level
                query_log_level = priv_query_exec.logging_level

        query_log_level = None
        query_exec = bt2.QueryExecutor(MySink, 'obj', None)
        query_exec.logging_level = bt2.LoggingLevel.WARNING
        query_exec.query()
        self.assertEqual(query_log_level, bt2.LoggingLevel.WARNING)
        del query_log_level

    def test_query_returns_none(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @staticmethod
            def _user_query(priv_query_exec, obj, params, method_obj):
                return

        res = bt2.QueryExecutor(MySink, 'obj', None).query()
        self.assertIsNone(res)

    def test_query_simple(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params, method_obj):
                nonlocal query_params
                query_params = params
                return 17.5

        query_params = None
        params = ['coucou', 23, None]
        res = bt2.QueryExecutor(MySink, 'obj', params).query()
        self.assertEqual(query_params, params)
        self.assertEqual(res, 17.5)
        del query_params

    def test_query_complex(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params, method_obj):
                nonlocal query_params
                query_params = params
                return {'null': None, 'bt2': 'BT2'}

        query_params = None
        params = {
            'array': ['coucou', 23, None],
            'other_map': {'yes': 'yeah', '19': 19, 'minus 1.5': -1.5},
            'null': None,
        }

        res = bt2.QueryExecutor(MySink, 'obj', params).query()
        self.assertEqual(query_params, params)
        self.assertEqual(res, {'null': None, 'bt2': 'BT2'})
        del query_params

    def test_eq(self):
        class MySink(bt2._UserSinkComponent):
            def _user_consume(self):
                pass

        self.assertEqual(MySink, MySink)


class ComponentClassTestCase(unittest.TestCase):
    def setUp(self):
        class MySink(bt2._UserSinkComponent):
            """
            The description.

            The help.
            """

            def _user_consume(self):
                pass

            @classmethod
            def _user_query(cls, priv_query_exec, obj, params, method_obj):
                return [obj, params, 23]

        self._py_comp_cls = MySink
        graph = bt2.Graph()
        comp = graph.add_component(MySink, 'salut')
        self._comp_cls = comp.cls
        self.assertIs(type(self._comp_cls), bt2._SinkComponentClassConst)

    def tearDown(self):
        del self._py_comp_cls
        del self._comp_cls

    def test_description(self):
        self.assertEqual(self._comp_cls.description, 'The description.')

    def test_help(self):
        self.assertEqual(self._comp_cls.help, 'The help.')

    def test_name(self):
        self.assertEqual(self._comp_cls.name, 'MySink')

    def test_addr(self):
        self.assertIsInstance(self._comp_cls.addr, int)
        self.assertNotEqual(self._comp_cls.addr, 0)

    def test_eq_invalid(self):
        self.assertFalse(self._comp_cls == 23)

    def test_eq(self):
        self.assertEqual(self._comp_cls, self._comp_cls)
        self.assertEqual(self._py_comp_cls, self._comp_cls)

    def test_query(self):
        res = bt2.QueryExecutor(
            self._comp_cls, 'an object', {'yes': 'no', 'book': -17}
        ).query()
        expected = ['an object', {'yes': 'no', 'book': -17}, 23]
        self.assertEqual(res, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016, Matt Layman

from io import StringIO
import re
import sys

from tap.directive import Directive
from tap.i18n import _
from tap.line import Bail, Diagnostic, Plan, Result, Unknown, Version


class Parser(object):
    """A parser for TAP files and lines."""

    # ok and not ok share most of the same characteristics.
    result_base = r"""
        \s*                    # Optional whitespace.
        (?P<number>\d*)        # Optional test number.
        \s*                    # Optional whitespace.
        (?P<description>[^#]*) # Optional description before #.
        \#?                    # Optional directive marker.
        \s*                    # Optional whitespace.
        (?P<directive>.*)      # Optional directive text.
    """
    ok = re.compile(r'^ok' + result_base, re.VERBOSE)
    not_ok = re.compile(r'^not\ ok' + result_base, re.VERBOSE)
    plan = re.compile(r"""
        ^1..(?P<expected>\d+) # Match the plan details.
        [^#]*                 # Consume any non-hash character to confirm only
                              # directives appear with the plan details.
        \#?                   # Optional directive marker.
        \s*                   # Optional whitespace.
        (?P<directive>.*)     # Optional directive text.
    """, re.VERBOSE)
    diagnostic = re.compile(r'^#')
    bail = re.compile(r"""
        ^Bail\ out!
        \s*            # Optional whitespace.
        (?P<reason>.*) # Optional reason.
    """, re.VERBOSE)
    version = re.compile(r'^TAP version (?P<version>\d+)$')

    TAP_MINIMUM_DECLARED_VERSION = 13

    def parse_file(self, filename):
        """Parse a TAP file to an iterable of tap.line.Line objects.

        This is a generator method that will yield an object for each
        parsed line. The file given by `filename` is assumed to exist.
        """
        return self.parse(open(filename, 'r'))

    def parse_stdin(self):
        """Parse a TAP stream from standard input.

        Note: this has the side effect of closing the standard input
        filehandle after parsing.
        """
        return self.parse(sys.stdin)

    def parse_text(self, text):
        """Parse a string containing one or more lines of TAP output."""
        return self.parse(StringIO(text))

    def parse(self, fh):
        """Generate tap.line.Line objects, given a file-like object `fh`.

        `fh` may be any object that implements both the iterator and
        context management protocol (i.e. it can be used in both a
        "with" statement and a "for...in" statement.)

        Trailing whitespace and newline characters will be automatically
        stripped from the input lines.
        """
        with fh:
            for line in fh:
                yield self.parse_line(line.rstrip())

    def parse_line(self, text):
        """Parse a line into whatever TAP category it belongs."""
        match = self.ok.match(text)
        if match:
            return self._parse_result(True, match)

        match = self.not_ok.match(text)
        if match:
            return self._parse_result(False, match)

        if self.diagnostic.match(text):
            return Diagnostic(text)

        match = self.plan.match(text)
        if match:
            return self._parse_plan(match)

        match = self.bail.match(text)
        if match:
            return Bail(match.group('reason'))

        match = self.version.match(text)
        if match:
            return self._parse_version(match)

        return Unknown()

    def _parse_plan(self, match):
        """Parse a matching plan line."""
        expected_tests = int(match.group('expected'))
        directive = Directive(match.group('directive'))

        # Only SKIP directives are allowed in the plan.
        if directive.text and not directive.skip:
            return Unknown()

        return Plan(expected_tests, directive)

    def _parse_result(self, ok, match):
        """Parse a matching result line into a result instance."""
        return Result(
            ok, match.group('number'), match.group('description').strip(),
            Directive(match.group('directive')))

    def _parse_version(self, match):
        version = int(match.group('version'))
        if version < self.TAP_MINIMUM_DECLARED_VERSION:
            raise ValueError(_('It is an error to explicitly specify '
                               'any version lower than 13.'))
        return Version(version)
